#! /usr/bin/env python

"""
Module with fake companion injection functions.
"""

__author__ = 'Carlos Alberto Gomez Gonzalez, Valentin Christiaens'
__all__ = ['collapse_psf_cube',
           'normalize_psf',
           'cube_inject_companions',
           'generate_cube_copies_with_injections',
           'frame_inject_companion']

import numpy as np
from scipy import stats
from scipy.interpolate import interp1d
import photutils
from ..preproc import (cube_crop_frames, frame_shift, frame_crop, cube_shift,
                       frame_rotate)
from ..var import (frame_center, fit_2dgaussian, fit_2dairydisk, fit_2dmoffat,
                   get_circle, get_annulus_segments)
from ..var.shapes import dist_matrix                   
from ..conf.utils_conf import print_precision, check_array


def cube_inject_companions(array, psf_template, angle_list, flevel, plsc,
                           rad_dists, n_branches=1, theta=0, imlib='opencv',
                           interpolation='lanczos4', transmission=None, 
                           full_output=False, verbose=True):
    """ Injects fake companions in branches, at given radial distances.

    Parameters
    ----------
    array : 3d/4d numpy ndarray
        Input cube. This is copied before the injections take place, so
        ``array`` is never modified.
    psf_template : numpy ndarray
        2d array with the normalized PSF template, with an odd or even shape.
        The PSF image must be centered wrt to the array! Therefore, it is
        recommended to run the function ``normalize_psf`` to generate a centered
        and flux-normalized PSF template. 
        It can also be a 3D array, but length should match ADI cube.
        In the ADI+mSDI (4D input cube) case it must be a 3d array.
    angle_list : 1d numpy ndarray
        List of parallactic angles, in degrees.
    flevel : float or list/1d array
        Factor for controlling the brightness of the fake companions. If a float, 
        the same flux is used for all frames. For a 3D input cube: if a 
        list/1d array is provided, it should have same length as number of 
        frames in the 3D cube. For a 4D (ADI+mSDI) input cube, a list/1d array 
        should have the same length as the number of spectral channels 
        (i.e. provide a spectrum).
    plsc : float
        Value of the plsc in arcsec/px. Only used for printing debug output when
        ``verbose=True``.
    rad_dists : float, list or array 1d
        Vector of radial distances of fake companions in pixels.
    n_branches : int, optional
        Number of azimutal branches.
    theta : float, optional
        Angle in degrees for rotating the position of the first branch that by
        default is located at zero degrees. Theta counts counterclockwise from
        the positive x axis.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    transmission: numpy array, optional
        Radial transmission of the coronagraph, if any. Array with 2 columns.
        First column is the radial separation in pixels. Second column is the
        corresponding off-axis transmission (between 0 and 1).
    full_output : bool, optional
        Returns the ``x`` and ``y`` coordinates of the injections, additionally
        to the new array.
    verbose : bool, optional
        If True prints out additional information.

    Returns
    -------
    array_out : numpy ndarray
        Output array with the injected fake companions.
    positions : list of tuple(y, x)
        [full_output] Coordinates of the injections in the first frame (and
        first wavelength for 4D cubes).
    psf_trans: array with injected psf affected by transmission (only returned
        if transmission is not None)
        

    """
    check_array(array, dim=(3, 4), msg="array")
    check_array(psf_template, dim=(2, 3), msg="psf_template")

    if array.ndim == 4 and psf_template.ndim != 3:
        raise ValueError('`psf_template` must be a 3d array')

    if not isinstance(plsc, float):
        raise TypeError("`plsc` must be a float")
    if not np.isscalar(flevel):
        if len(flevel) != array.shape[0]:
            msg = "if not scalar `flevel` must have same length as array"
            raise TypeError(msg) 

    ## set imlib for rotation & shift (rotation used if transmission!=None)
    if imlib == 'opencv':
        imlib_sh = imlib
        imlib_rot = imlib
    elif imlib == 'skimage' or imlib == 'ndimage-interp':
        imlib_sh = 'ndimage-interp'
        imlib_rot = 'skimage'
    elif imlib == 'vip-fft' or imlib == 'ndimage-fourier':
        imlib_sh = 'ndimage-fourier'
        imlib_rot = 'vip-fft'
    else:
        raise TypeError("Interpolation not recognized.")

    rad_dists = np.asarray(rad_dists).reshape(-1)  # forces ndim=1
    positions = []

    if transmission is not None:
        if transmission.shape[0] != 2:
            raise ValueError("transmission should be a (2,N) ndarray")
        # if transmission doesn't have right format for interpolation, adapt it
        if transmission[0,0] != 0 or transmission[0,-1] < np.sqrt(2)*array.shape[-1]:
            trans_rad_list = transmission[0].tolist()
            trans_list = transmission[1].tolist()
            ## should have a zero point        
            if transmission[0,0] != 0:
                trans_rad_list = [0]+trans_rad_list
                trans_list = [0]+trans_list
            ## last point should be max possible distance between fc and star
            if transmission[0,-1] < np.sqrt(2)*array.shape[-1]:
                trans_rad_list = trans_rad_list+[np.sqrt(2)*array.shape[-1]]
                trans_list = trans_list+[1]
            transmission = np.array([trans_rad_list,trans_list])
        
        ## last radial separation should be beyond the edge of frame
        interp_trans = interp1d(transmission[0],transmission[1])

    # ADI case
    if array.ndim == 3:
        ceny, cenx = frame_center(array[0])

        if not rad_dists[-1] < array[0].shape[0] / 2:
            raise ValueError('rad_dists last location is at the border (or '
                             'outside) of the field')
        size_fc = psf_template.shape[1]
        nframes = array.shape[0]

        w = int(np.ceil(size_fc/2)) - 1
        sty = int(ceny) - w
        sty = int(cenx) - w

        # fake companion in the center of a zeros frame
        fc_fr = np.zeros_like(array)
        if psf_template.ndim == 2:
            try:
                for fr in range(nframes):
                    fc_fr[fr, sty:sty+size_fc, sty:sty+size_fc] = psf_template
            except ValueError as e:
                print("cannot place PSF on frame. Please verify the shapes! "
                      "psf shape: {}, array shape: {}".format(psf_template.shape,
                                                              array.shape))
                raise e
        else:
            try:
                for fr in range(nframes):
                    fc_fr[fr,sty:sty+size_fc, sty:sty+size_fc] = psf_template[fr]
            except ValueError as e:
                print("cannot place PSF on frames. Please verify the shapes!"
                      "psf shape: {}, array shape: {}".format(psf_template.shape,
                                                              array.shape))
                raise e

        array_out = array.copy()

        for branch in range(n_branches):
            ang = (branch * 2 * np.pi / n_branches) + np.deg2rad(theta)

            if verbose:
                print('Branch {}:'.format(branch+1))                

            for rad in rad_dists:
                if transmission is not None:
                    fc_fr_rad = fc_fr.copy()
                    y_star = ceny
                    x_star = cenx - rad
                    d = dist_matrix(fc_fr.shape[1],x_star,y_star)
                    for i in range(d.shape[0]):
                        fc_fr_rad[:,i] = interp_trans(d[i])*fc_fr[:,i]
                    if full_output:
                        # check the effect of transmission on a single PSF tmp
                        psf_trans = frame_rotate(fc_fr_rad[0],
                                                 -(np.rad2deg(ang)),
                                                 imlib=imlib_rot,
                                                 interpolation=interpolation)
                        shift_y = rad * np.sin(ang)
                        shift_x = rad * np.cos(ang)
                        psf_trans = frame_shift(psf_trans, shift_y, shift_x,
                                                imlib_sh, interpolation)
                    
                for fr in range(nframes):
                    shift_y = rad * np.sin(ang - np.deg2rad(angle_list[fr]))
                    shift_x = rad * np.cos(ang - np.deg2rad(angle_list[fr]))
                    if transmission is not None:                       
                        fc_fr_ang = frame_rotate(fc_fr_rad[fr], 
                                                 -(ang*180/np.pi-angle_list[fr]),
                                                 imlib=imlib_rot,
                                                 interpolation=interpolation)
                    else:
                        fc_fr_ang = fc_fr[fr]
                    if np.isscalar(flevel):
                        array_out[fr] += (frame_shift(fc_fr_ang, shift_y, 
                                                      shift_x, imlib_sh, 
                                                      interpolation)
                                      * flevel)
                    else:
                        array_out[fr] += (frame_shift(fc_fr_ang, shift_y, 
                                                      shift_x, imlib_sh, 
                                                      interpolation)
                                  * flevel[fr])
                            
                pos_y = rad * np.sin(ang) + ceny
                pos_x = rad * np.cos(ang) + cenx
                rad_arcs = rad * plsc

                positions.append((pos_y, pos_x))

                if verbose:
                    print('\t(X,Y)=({:.2f}, {:.2f}) at {:.2f} arcsec '
                          '({:.2f} pxs from center)'.format(pos_x, pos_y,
                                                            rad_arcs, rad))

    # ADI+mSDI (IFS) case
    if array.ndim == 4 and psf_template.ndim == 3:
        ceny, cenx = frame_center(array[0, 0])

        if not rad_dists[-1] < array[0,0].shape[0] / 2:
            raise ValueError('rad_dists last location is at the border (or '
                             'outside) of the field')

        sizey = array.shape[2]
        sizex = array.shape[3]
        size_fc = psf_template.shape[2]  # considering square frames
        nframes_wav = array.shape[0]
        nframes_adi = array.shape[1]
        fc_fr = np.zeros((nframes_wav, sizey, sizex), dtype=np.float64) # -> 3d

        for i in range(nframes_wav):
            w = int(np.floor(size_fc/2.))
            # fake companion in the center of a zeros frame
            if (psf_template[0].shape[1] % 2) == 0:
                fc_fr[i, int(ceny)-w:int(ceny)+w,
                      int(cenx)-w:int(cenx)+w] = psf_template[i]
            else:
                fc_fr[i, int(ceny)-w:int(ceny)+w+1,
                      int(cenx)-w:int(cenx)+w+1] = psf_template[i]

        array_out = array.copy()

        for branch in range(n_branches):
            ang = (branch * 2 * np.pi / n_branches) + np.deg2rad(theta)

            if verbose:
                print('Branch {}:'.format(branch+1))

            for rad in rad_dists:

                for fr in range(nframes_adi):
                    shift_y = rad * np.sin(ang - np.deg2rad(angle_list[fr]))
                    shift_x = rad * np.cos(ang - np.deg2rad(angle_list[fr]))
                    shift = cube_shift(fc_fr, shift_y, shift_x, imlib_sh,
                                       interpolation)

                    if np.isscalar(flevel):
                        array_out[:, fr] += shift * flevel
                    else:
                        for i in range(len(flevel)):
                            array_out[i, fr] += shift[i] * flevel[i]

                pos_y = rad * np.sin(ang) + ceny
                pos_x = rad * np.cos(ang) + cenx
                rad_arcs = rad * plsc

                positions.append((pos_y, pos_x))

                if verbose:
                    print('\t(X,Y)=({:.2f}, {:.2f}) at {:.2f} arcsec '
                          '({:.2f} pxs from center)'.format(pos_x, pos_y,
                                                            rad_arcs, rad))

    if full_output:
        if transmission is not None:
            return array_out, positions, psf_trans
        else:
            return array_out, positions
    else:
        return array_out


def generate_cube_copies_with_injections(array, psf_template, angle_list, plsc,
                                         n_copies=100, inrad=8, outrad=12,
                                         dist_flux=("uniform", 2, 500)):
    """
    Create multiple copies of ``array`` with different random injections.

    This is a wrapper around ``metrics.cube_inject_companions``, which deals
    with multiple copies of the original data cube and generates random
    parameters.

    Parameters
    ----------
    array : 3d/4d numpy ndarray
        Original input cube.
    psf_template : 2d/3d numpy ndarray
        Array with the normalized psf template. It should have an odd shape.
        It's recommended to run the function ``normalize_psf`` to get a proper
        PSF template. In the ADI+mSDI case it must be a 3d array.
    angle_list : 1d numpy ndarray
        List of parallactic angles, in degrees.
    plsc : float
        Value of the plsc in arcsec/px. Only used for printing debug output when
        ``verbose=True``.
    n_copies : int
        This is the number of 'cube copies' returned.
    inrad,outrad : float
        Inner and outer radius of the injections. The actual injection position
        is chosen randomly.
    dist_flux : tuple('method', *params)
        Tuple describing the flux selection. Method can be a function, the
        ``*params`` are passed to it. Method can also be a string, for a
        pre-defined random function:

            ``("skewnormal", skew, mean, var)``
                uses scipy.stats.skewnorm.rvs
            ``("uniform", low, high)``
                uses np.random.uniform
            ``("normal", loc, scale)``
                uses np.random.normal

    Yields
    ------
    fake_data : dict
        Represents a copy of the original ``array``, with fake injections. The
        dictionary keys are:

            ``cube``
                Array shaped like the input ``array``, with the fake injections.
            ``position`` : list of tuples(y,x)
                List containing the positions of the injected companions, as
                (y,x) tuples.
            ``dist`` : float
                The distance of the injected companions, which was passed to
                ``cube_inject_companions``.
            ``theta`` : float, degrees
                The initial angle, as passed to ``cube_inject_companions``.
            ``flux`` : float
                The flux passed to ``cube_inject_companions``.

    """
    # TODO: 'mask' parameter for known companions?

    width = outrad - inrad
    yy, xx = get_annulus_segments(array[0], inrad, width)[0]
    num_patches = yy.shape[0]

    # Defining Fluxes according to chosen distribution
    dist_fkt = dict(skewnormal=stats.skewnorm.rvs,
                    normal=np.random.normal,
                    uniform=np.random.uniform).get(dist_flux[0],
                                                   dist_flux[0])
    fluxes = sorted(dist_fkt(*dist_flux[1:], size=n_copies))

    inds_inj = np.random.randint(0, num_patches, size=n_copies)

    # Injections
    for n in range(n_copies):

        injx = xx[inds_inj[n]] - frame_center(array[0])[1]
        injy = yy[inds_inj[n]] - frame_center(array[0])[0]
        dist = np.sqrt(injx**2 + injy**2)
        theta = np.mod(np.arctan2(injy, injx) / np.pi * 180, 360)

        fake_cube, positions = cube_inject_companions(
            array, psf_template, angle_list, plsc=plsc,
            flevel=fluxes[n], theta=theta,
            rad_dists=dist, n_branches=1,  # TODO: multiple injections?
            full_output=True, verbose=False
        )

        yield dict(
            positions=positions,
            dist=dist, theta=theta, flux=fluxes[n],
            cube=fake_cube
        )


def frame_inject_companion(array, array_fc, pos_y, pos_x, flux,
                           imlib='opencv', interpolation='lanczos4'):
    """ Injects a fake companion in a single frame (it could be a single
     multi-wavelength frame) at given coordinates.
    """
    if not (array.ndim == 2 or array.ndim == 3):
        raise TypeError('Array is not a 2d or 3d array.')
    if array.ndim == 2:
        size_fc = array_fc.shape[0]
        ceny, cenx = frame_center(array)
        ceny = int(ceny)
        cenx = int(cenx)
        fc_fr = np.zeros_like(array)
        w = int(np.floor(size_fc/2.))
        # fake companion in the center of a zeros frame
        fc_fr[ceny-w:ceny+w+1, cenx-w:cenx+w+1] = array_fc
        array_out = array + frame_shift(fc_fr, pos_y-ceny, pos_x-cenx, imlib,
                                        interpolation) * flux

    if array.ndim == 3:
        size_fc = array_fc.shape[1]
        ceny, cenx = frame_center(array[0])
        ceny = int(ceny)
        cenx = int(cenx)
        fc_fr = np.zeros_like(array)
        w = int(np.floor(size_fc/2.))
        # fake companion in the center of a zeros frame
        fc_fr[:, ceny-w:ceny+w+1, cenx-w:cenx+w+1] = array_fc
        array_out = array + cube_shift(fc_fr, pos_y - ceny, pos_x - cenx,
                                       imlib, interpolation) * flux

    return array_out


def collapse_psf_cube(array, size, fwhm=4, verbose=True, collapse='mean'):
    """ Creates a 2d PSF template from a cube of non-saturated off-axis frames
    of the star by taking the mean and normalizing the PSF flux.

    Parameters
    ----------
    array : numpy ndarray, 3d
        Input cube.
    size : int
        Size of the squared subimage.
    fwhm: float, optional
        The size of the Full Width Half Maximum in pixel.
    verbose : {True,False}, bool optional
        Whether to print to stdout information about file opening, cropping and
        completion of the psf template.
    collapse : {'mean','median'}, string optional
        Defines the way the frames are collapsed.

    Returns
    -------
    psf_normd : numpy ndarray
        Normalized PSF.
    """
    if array.ndim != 3 and array.ndim != 4:
        raise TypeError('Array is not a cube, 3d or 4d array.')

    n = array.shape[0]
    psf = cube_crop_frames(array, size=size, verbose=verbose)
    if collapse == 'mean':
        psf = np.mean(psf, axis=0)
    elif collapse == 'median':
        psf = np.median(psf, axis=0)
    else:
        raise TypeError('Collapse mode not recognized.')

    psf_normd = normalize_psf(psf, size=size, fwhm=fwhm)

    if verbose:
        print("Done scaled PSF template from the average of", n, "frames.")
    return psf_normd


def normalize_psf(array, fwhm='fit', size=None, threshold=None, mask_core=None,
                  model='gauss', imlib='opencv', interpolation='lanczos4',
                  force_odd=True, full_output=False, verbose=True, debug=False):
    """ Normalizes a PSF (2d or 3d array), to have the flux in a 1xFWHM
    aperture equal to one. It also allows to crop the array and center the PSF
    at the center of the array(s).

    Parameters
    ----------
    array: numpy ndarray
        The PSF, 2d (ADI data) or 3d array (IFS data).
    fwhm: int, float, 1d array or str, optional
        The the Full Width Half Maximum in pixels. It can handle a different
        FWHM value for different wavelengths (IFS data). If set to 'fit' then
        a ``model`` (assuming the PSF is centered in the array) is fitted to
        estimate the FWHM in 2D or 3D PSF arrays.
    size : int or None, optional
        If int it will correspond to the size of the centered sub-image to be
        cropped form the PSF array. The PSF is assumed to be rougly centered wrt
        the array.
    threshold : None of float, optional
        Sets to zero small values, trying to leave only the core of the PSF.
    mask_core : None of float, optional
        Sets the radius of a circular aperture for the core of the PSF,
        everything else will be set to zero.
    model : {'gauss', 'moff', 'airy'}, str optional
        The assumed model used to fit the PSF: either a Gaussian, a Moffat
        or an Airy 2d distribution.
    imlib : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    interpolation : str, optional
        See the documentation of the ``vip_hci.preproc.frame_shift`` function.
    force_odd : str, optional
        If True the resulting array will have odd size (and the PSF will be
        placed at its center). If False, and the frame size is even, then the
        PSF will be put at the center of an even-sized frame.
    full_output : bool, optional
        If True the flux in a FWHM aperture is returned along with the
        normalized PSF.
    verbose : bool, optional
        If True intermediate results are printed out.
    debug : bool, optional
        If True the fitting will output additional information and a diagnostic
        plot will be shown (this might cause a long output if ``array`` is 3d
        and has many slices).

    Returns
    -------
    psf_norm : numpy ndarray
        The normalized PSF (2d or 3d array).
    fwhm_flux : numpy ndarray
        [full_output=True] The flux in a FWHM aperture (it can be a single
        value or a vector).
    fwhm : numpy ndarray
        [full_output=True] The FWHM size. If ``fwhm`` is set to 'fit' then it
        is the fitted FWHM value according to the assumed ``model`` (the mean in
        X and Y is returned when ``model`` is set to 'gauss').

    """
    def psf_norm_2d(psf, fwhm, threshold, mask_core, full_output, verbose):
        """ 2d case """
        # we check if the psf is centered and fix it if needed
        cy, cx = frame_center(psf, verbose=False)
        xcom, ycom = photutils.centroid_com(psf)
        if not (np.allclose(cy, ycom, atol=1e-2) or
                np.allclose(cx, xcom, atol=1e-2)):
            # first we find the centroid and put it in the center of the array
            centry, centrx = fit_2d(psf, full_output=False, debug=False)
            shiftx, shifty = centrx - cx, centry - cy
            psf = frame_shift(psf, -shifty, -shiftx, imlib=imlib,
                              interpolation=interpolation)

            for _ in range(2):
                centry, centrx = fit_2d(psf, full_output=False, debug=False)
                cy, cx = frame_center(psf, verbose=False)
                shiftx, shifty = centrx - cx, centry - cy
                psf = frame_shift(psf, -shifty, -shiftx, imlib=imlib,
                                  interpolation=interpolation)

        # we check whether the flux is normalized and fix it if needed
        fwhm_aper = photutils.CircularAperture((cx,cy), fwhm/2)
        fwhm_aper_phot = photutils.aperture_photometry(psf, fwhm_aper,
                                                       method='exact')
        fwhm_flux = np.array(fwhm_aper_phot['aperture_sum'])

        if fwhm_flux > 1.1 or fwhm_flux < 0.9:
            psf_norm_array = psf / np.array(fwhm_aper_phot['aperture_sum'])
        else:
            psf_norm_array = psf

        if threshold is not None:
            psf_norm_array[np.where(psf_norm_array < threshold)] = 0

        if mask_core is not None:
            psf_norm_array = get_circle(psf_norm_array, radius=mask_core)

        if verbose:
            print("Flux in 1xFWHM aperture: {:.3f}".format(fwhm_flux[0]))

        if full_output:
            return psf_norm_array, fwhm_flux, fwhm
        else:
            return psf_norm_array
    ###########################################################################
    if model == 'gauss':
        fit_2d = fit_2dgaussian
    elif model == 'moff':
        fit_2d = fit_2dmoffat
    elif model == 'airy':
        fit_2d = fit_2dairydisk
    else:
        raise ValueError('`Model` not recognized')

    if array.ndim == 2:
        y, x = array.shape
        if size is not None:
            if force_odd and size % 2 == 0:
                size += 1
                msg = "`Force_odd` is True therefore `size` was set to {}"
                print(msg.format(size))
        else:
            if force_odd and y % 2 == 0:
                size = y - 1
                msg = "`Force_odd` is True and frame size is even, therefore "
                msg += "new frame size was set to {}"
                print(msg.format(size))

        if size is not None:
            if size < array.shape[0]:
                array = frame_crop(array, size, force=True, verbose=False)
            else:
                array = array.copy()
        else:
            array = array.copy()

        if fwhm == 'fit':
            fit = fit_2d(array, full_output=True, debug=debug)
            if model == 'gauss':
                fwhm = np.mean((fit['fwhm_x'], fit['fwhm_y']))
                if verbose:
                    print("\nMean FWHM: {:.3f}".format(fwhm))
            elif model == 'moff' or model == 'airy':
                fwhm = fit.fwhm.at[0]
                if verbose:
                    print("FWHM: {:.3f}".format(fwhm))

        res = psf_norm_2d(array, fwhm, threshold, mask_core, full_output,
                          verbose)
        return res

    elif array.ndim == 3:
        n, y, x = array.shape
        if size is not None:
            if force_odd and size % 2 == 0:
                size += 1
                msg = "`Force_odd` is True therefore `size` was set to {}"
                print(msg.format(size))
        else:
            if force_odd and y % 2 == 0:
                size = y - 1
                msg = "`Force_odd` is True and frame size is even, therefore "
                msg += "new frame size was set to {}"
                print(msg.format(size))

        if size is not None:
            if size < array.shape[1]:
                array = cube_crop_frames(array, size, force=True, verbose=False)
            else:
                array = array.copy()

        if isinstance(fwhm, (int, float)):
            fwhm = [fwhm] * array.shape[0]
        elif fwhm == 'fit':
            fits_vect = [fit_2d(array[i], full_output=True, debug=debug) for i
                         in range(n)]
            if model == 'gauss':
                fwhmx = [fits_vect[i]['fwhm_x'] for i in range(n)]
                fwhmy = [fits_vect[i]['fwhm_y'] for i in range(n)]
                fwhm_vect = [np.mean((fwhmx[i], fwhmy[i])) for i in range(n)]
                fwhm = np.array(fwhm_vect)
                if verbose:
                    print("Mean FWHM per channel: ")
                    print_precision(fwhm)
            elif model == 'moff' or model == 'airy':
                fwhm_vect = [fits_vect[i]['fwhm'] for i in range(n)]
                fwhm = np.array(fwhm_vect)
                fwhm = fwhm.flatten()
                if verbose:
                    print("FWHM per channel:")
                    print_precision(fwhm)

        array_out = []
        fwhm_flux = np.zeros(n)

        for fr in range(array.shape[0]):
            restemp = psf_norm_2d(array[fr], fwhm[fr], threshold, mask_core,
                                  True, False)
            array_out.append(restemp[0])
            fwhm_flux[fr] = restemp[1]

        array_out = np.array(array_out)
        if verbose:
            print("Flux in 1xFWHM aperture: ")
            print_precision(fwhm_flux)

        if full_output:
            return array_out, fwhm_flux, fwhm
        else:
            return array_out



#! /usr/bin/env python

"""
Module with utility functions to the MCMC (``emcee``) sampling for NEGFC 
parameter estimation.
"""


__author__ = 'V. Christiaens, O. Wertz, Carlos Alberto Gomez Gonzalez'
__all__ = ['gelman_rubin',
           'gelman_rubin_from_chain',
           'autocorr_func_1d',
           'auto_window',
           'autocorr',
           'autocorr_test']

import numpy as np
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)


def gelman_rubin(x):
    """
    Determine the Gelman-Rubin \hat{R} statistical test between Markov chains.
    
    Parameters
    ----------
    x: numpy.array
        The numpy.array on which the Gelman-Rubin test is applied. This array
        should contain at least 2 set of data, i.e. x.shape >= (2,).
        
    Returns
    -------
    out: float
        The Gelman-Rubin \hat{R}.

    Example
    -------
    >>> x1 = np.random.normal(0.0,1.0,(1,100))
    >>> x2 = np.random.normal(0.1,1.3,(1,100))
    >>> x = np.vstack((x1,x2))
    >>> gelman_rubin(x)
    1.0366629898991262
    >>> gelman_rubin(np.vstack((x1,x1)))
    0.99
        
    """
    if np.shape(x) < (2,):
        msg = 'Gelman-Rubin diagnostic requires multiple chains of the same '
        msg += 'length'
        raise ValueError(msg)

    try:
        m, n = np.shape(x)
    except ValueError:
        print("Bad shape for the chains")
        return

    # Calculate between-chain variance
    B_over_n = np.sum((np.mean(x, 1) - np.mean(x)) ** 2) / (m - 1)

    # Calculate within-chain variances
    W = np.sum([(x[i] - xbar) ** 2 for i, xbar in
                enumerate(np.mean(x, 1))]) / (m * (n - 1))
    # (over) estimate of variance
    s2 = W * (n - 1) / n + B_over_n

    # Pooled posterior variance estimate
    V = s2 + B_over_n / m

    # Calculate PSRF
    R = V / W

    return R


def gelman_rubin_from_chain(chain, burnin):
    """ Pack the MCMC chain and determine the Gelman-Rubin \hat{R} statistical
    test. In other words, two sub-sets are extracted from the chain (burnin
    parts are taken into account) and the Gelman-Rubin statistical test is
    performed.
    
    Parameters
    ----------
    chain: numpy.array
        The MCMC chain with the shape walkers x steps x model_parameters
    burnin: float \in [0,1]
        The fraction of a walker which is discarded.
        
    Returns
    -------
    out: float
        The Gelman-Rubin \hat{R}.
        
    """
    dim = chain.shape[2]
    k = chain.shape[1]
    thr0 = int(np.floor(burnin*k))
    thr1 = int(np.floor((1-burnin) * k * 0.25))
    rhat = np.zeros(dim)
    for j in range(dim):
        part1 = chain[:, thr0:thr0+thr1, j].reshape((-1))
        part2 = chain[:, thr0+3*thr1:thr0+4*thr1, j].reshape((-1))
        series = np.vstack((part1, part2))
        rhat[j] = gelman_rubin(series)
    return rhat


def next_pow_two(n):
    i = 1
    while i < n:
        i = i << 1
    return i


def autocorr_func_1d(x, norm=True):
    x = np.atleast_1d(x)
    if len(x.shape) != 1:
        raise ValueError("invalid dimensions for 1D autocorrelation function")
    n = next_pow_two(len(x))

    # Compute the FFT and then (from that) the auto-correlation function
    f = np.fft.fft(x - np.mean(x), n=2 * n)
    acf = np.fft.ifft(f * np.conjugate(f))[: len(x)].real
    acf /= 4 * n

    # Optionally normalize
    if norm:
        acf /= acf[0]

    return acf


def auto_window(taus, c):
    m = np.arange(len(taus)) < c * taus
    if np.any(m):
        return np.argmin(m)
    return len(taus) - 1


def autocorr(y, c=5.0):
    f = np.zeros(y.shape[1])
    for yy in y:
        f += autocorr_func_1d(yy)
    f /= len(y)
    taus = 2.0 * np.cumsum(f) - 1.0
    window = auto_window(taus, c)
    return taus[window]


def autocorr_test(chain):
    N = chain.shape[1]
    tau = autocorr(chain)
    return tau/N


#! /usr/bin/env python

"""
Module with functions for computing SVDs.
"""

__author__ = 'Carlos Alberto Gomez Gonzalez'
__all__ = ['SVDecomposer']

import warnings
try:
    import cupy
    no_cupy = False
except ImportError:
    msg = "Cupy not found. Do you have a GPU? Consider setting up a CUDA "
    msg += "environment and installing cupy >= 2.0.0"
    warnings.warn(msg, ImportWarning)
    no_cupy = True
try:
    import torch
    no_torch = False
except ImportError:
    msg = "Pytorch not found. Do you have a GPU? Consider setting up a CUDA "
    msg += "environment and installing pytorch"
    warnings.warn(msg, ImportWarning)
    no_torch = True

import numpy as np
from numpy import linalg
from matplotlib import pyplot as plt 
from scipy.sparse.linalg import svds
from sklearn.decomposition import randomized_svd
from sklearn.utils import check_random_state
from pandas import DataFrame
from ..conf import timing, time_ini, sep, Progressbar
from ..var import matrix_scaling, prepare_matrix
from ..preproc import check_scal_vector, cube_crop_frames
from ..preproc import cube_rescaling_wavelengths as scwave
from ..conf import vip_figsize, check_array


class SVDecomposer:
    """
    Class for SVD decomposition of 2d, 3d or 4d HCI arrays.

    Parameters
    ----------
    data : numpy ndarray
        Input array (2d, 3d or 4d).
    mode : {'fullfr', 'annular'}, optional
        Whether to use the whole frames or a single annulus.
    inrad : None or int, optional
        [mode='annular'] Inner radius.
    outrad : None or int, optional
        [mode='annular'] Outer radius.
    svd_mode : {'lapack', 'arpack', 'eigen', 'randsvd', 'cupy', 'eigencupy',
        'randcupy', 'pytorch', 'eigenpytorch', 'randpytorch'}, str optional
        Switch for the SVD method/library to be used.

        ``lapack``: uses the LAPACK linear algebra library through Numpy
        and it is the most conventional way of computing the SVD
        (deterministic result computed on CPU).

        ``arpack``: uses the ARPACK Fortran libraries accessible through
        Scipy (computation on CPU).

        ``eigen``: computes the singular vectors through the
        eigendecomposition of the covariance M.M' (computation on CPU).

        ``randsvd``: uses the randomized_svd algorithm implemented in
        Sklearn (computation on CPU).

        ``cupy``: uses the Cupy library for GPU computation of the SVD as in
        the LAPACK version. `

        `eigencupy``: offers the same method as with the ``eigen`` option
        but on GPU (through Cupy).

        ``randcupy``: is an adaptation of the randomized_svd algorithm,
        where all the computations are done on a GPU (through Cupy). `

        `pytorch``: uses the Pytorch library for GPU computation of the SVD.

        ``eigenpytorch``: offers the same method as with the ``eigen``
        option but on GPU (through Pytorch).

        ``randpytorch``: is an adaptation of the randomized_svd algorithm,
        where all the linear algebra computations are done on a GPU
        (through Pytorch).

    scaling : {None, "temp-mean", spat-mean", "temp-standard",
               "spat-standard"}, None or str optional
        Pixel-wise scaling mode using ``sklearn.preprocessing.scale``
        function. If set to None, the input matrix is left untouched.
        Otherwise:

        ``temp-mean``: temporal px-wise mean is subtracted.

        ``spat-mean``: spatial mean is subtracted.

        ``temp-standard``: temporal mean centering plus scaling pixel values
        to unit variance.

        ``spat-standard``: spatial mean centering plus scaling pixel values
        to unit variance.

    wavelengths : numpy ndarray, optional
        Wavelengths in case of a 4d HCI cube. These are used to compute
        scaling factors for re-scaling the spectral channels and aligning
        the speckles.

    verbose : bool, optional
        If True intermediate messages and timing are printed.

    Notes
    -----
    For info on CEVR search: # Get variance explained by singular values in
    https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/decomposition/pca.py
    """
    def __init__(self, data, mode='fullfr', inrad=None, outrad=None,
                 svd_mode='lapack', scaling='temp-standard', wavelengths=None,
                 verbose=True):
        """
        """
        check_array(data, (2, 3, 4), msg='data')
        self.data = data
        self.mode = mode
        self.svd_mode = svd_mode
        self.inrad = inrad
        self.outrad = outrad
        self.scaling = scaling
        self.wavelengths = wavelengths
        self.verbose = verbose

        if self.mode == 'annular':
            if inrad is None:
                raise ValueError("`inrad` must be a positive integer")
            if outrad is None:
                raise ValueError("`outrad` must be a positive integer")

        if self.verbose:
            print(sep)

    def generate_matrix(self):
        """
        Generate a matrix from the input ``data``. Pixel values in the matrix
        are scaled. Depending on ``mode``, the matrix can come from an annulus
        instead of the whole frames.
        """
        start_time = time_ini(False)
        if self.data.ndim == 2:
            print("`data` is already a 2d array")
            self.matrix = matrix_scaling(self.data, self.scaling)

        elif self.data.ndim in [3, 4]:
            if self.data.ndim == 3:
                cube_ = self.data

            elif self.data.ndim == 4:
                if self.wavelengths is None:
                    raise ValueError("`wavelengths` must be provided when "
                                     "`data` is a 4D array")
                z, n_frames, y_in, x_in = self.data.shape
                scale_list = check_scal_vector(self.wavelengths)
                if not scale_list.shape[0] == z:
                    raise ValueError("`wavelengths` length is {} instead of "
                                     "{}".format(scale_list.shape[0], z))
                big_cube = []
                # Rescaling the spectral channels to align the speckles
                if self.verbose:
                    print('Rescaling the spectral channels to align the '
                          'speckles')
                for i in Progressbar(range(n_frames), verbose=self.verbose):
                    cube_resc = scwave(self.data[:, i, :, :], scale_list)[0]
                    cube_resc = cube_crop_frames(cube_resc, size=y_in,
                                                 verbose=False)
                    big_cube.append(cube_resc)
                big_cube = np.array(big_cube)
                cube_ = big_cube.reshape(z * n_frames, y_in, x_in)
                self.cube4dto3d_shape = cube_.shape

            result = prepare_matrix(cube_, self.scaling, mode=self.mode,
                                    inner_radius=self.inrad,
                                    outer_radius=self.outrad,
                                    verbose=self.verbose)
            if self.mode == 'annular':
                self.matrix = result[0]
                pxind = result[1]
                self.yy, self.xx = pxind  # pixel coords in the annulus
            elif self.mode == 'fullfr':
                self.matrix = result

        if self.verbose:
            timing(start_time)

    def run(self):
        """
        Decompose the input data.
        """
        start_time = time_ini(False)
        if not hasattr(self, 'matrix'):
            self.generate_matrix()

        max_pcs = min(self.matrix.shape[0], self.matrix.shape[1])

        results = svd_wrapper(self.matrix, self.svd_mode, max_pcs,
                              verbose=self.verbose, full_output=True)
        if len(results) == 3:
            self.u, self.s, self.v = results
        elif len(results) == 2:
            self.s, self.v = results

        if self.verbose:
            timing(start_time)

    def get_cevr(self, ncomp_list=None, plot=True, plot_save=False, plot_dpi=90,
                 plot_truncation=None):
        """
        Calculate the cumulative explained variance ratio for the SVD of a
        cube/matrix (either full frames or a single annulus could be used).

        Parameters
        ----------
        ncomp_list : None, list or tuple, optional
            If provided the list is used to filter the vector of CEVR.
        plot : bool, optional
            If True, the CEVR is plotted.
        plot_save : bool, optional
            If True, the plot is saved as ./figure.pdf.
        plot_dpi : int, optional
            The DPI of the figure.
        plot_truncation : None or int, optional
            If provided, it created a second panel in the plot, focusing on the
            CEVR curve up to ``plot_truncation`` components.

        Returns
        -------
        df_allks : pandas dataframe
            [ncomp_list is None] A table with the explained varaince ratio and
            the CEVR for all ncomps.
        df_klist : pandas dataframe
            [ncomp_list is not None] A table with the ncomp_list, the explained
            varaince ratio and the CEVR.
        """
        start_time = time_ini(False)
        if not hasattr(self, 'v'):
            self.run()

        if self.verbose:
            print("Computing the cumulative explained variance ratios")

        self.ncomp_list = ncomp_list
        exp_var = (self.s ** 2) / (self.s.shape[0] - 1)
        full_var = np.sum(exp_var)
        # % of variance explained by each PC
        self.explained_variance_ratio = exp_var / full_var
        self.cevr = np.cumsum(self.explained_variance_ratio)

        df_allks = DataFrame({'ncomp': range(1, self.s.shape[0] + 1),
                              'expvar_ratio': self.explained_variance_ratio,
                              'cevr': self.cevr})
        self.table_cevr = df_allks

        if plot:
            lw = 2;
            alpha = 0.4
            fig = plt.figure(figsize=vip_figsize, dpi=plot_dpi)
            fig.subplots_adjust(wspace=0.4)
            ax1 = plt.subplot2grid((1, 3), (0, 0), colspan=2)
            ax1.step(range(self.explained_variance_ratio.shape[0]),
                     self.explained_variance_ratio, alpha=alpha, where='mid',
                     label='Individual EVR', lw=lw)
            ax1.plot(self.cevr, '.-', alpha=alpha,
                     label='Cumulative EVR', lw=lw)
            ax1.legend(loc='best', frameon=False, fontsize='medium')
            ax1.set_ylabel('Explained variance ratio (EVR)')
            ax1.set_xlabel('Principal components')
            ax1.grid(linestyle='solid', alpha=0.2)
            ax1.set_xlim(-10, self.explained_variance_ratio.shape[0] + 10)
            ax1.set_ylim(0, 1)

            if plot_truncation is not None:
                ax2 = plt.subplot2grid((1, 3), (0, 2), colspan=1)
                ax2.step(range(plot_truncation),
                         self.explained_variance_ratio[:plot_truncation],
                         alpha=alpha, where='mid', lw=lw)
                ax2.plot(self.cevr[:plot_truncation], '.-', alpha=alpha, lw=lw)
                ax2.set_xlabel('Principal components')
                ax2.grid(linestyle='solid', alpha=0.2)
                ax2.set_xlim(-2, plot_truncation + 2)
                ax2.set_ylim(0, 1)

            if plot_save:
                plt.savefig('figure.pdf', dpi=300, bbox_inches='tight')

        if self.ncomp_list is not None:
            cevr_klist = []
            expvar_ratio_klist = []
            for k in self.ncomp_list:
                cevr_klist.append(self.cevr[k - 1])
                expvar_ratio_klist.append(self.explained_variance_ratio[k - 1])

            df_klist = DataFrame({'ncomp': self.ncomp_list,
                                  'exp_var_ratio': expvar_ratio_klist,
                                  'cevr': cevr_klist})
            self.cevr_ncomp = cevr_klist
            self.table_cevr_ncomp = df_klist
            if self.verbose:
                timing(start_time)
            return df_klist
        else:
            if self.verbose:
                timing(start_time)
            return df_allks

    def cevr_to_ncomp(self, cevr=0.9):
        """
        Infer the number of principal components for a given CEVR.

        Parameters
        ----------
        cevr : float or tuple of floats, optional
            The desired CEVR.

        Returns
        -------
        ncomp : int or list of ints
            The found number(s) of PCs.

        """
        if not hasattr(self, 'cevr'):
            self.get_cevr(plot=False)

        if isinstance(cevr, float):
            ncomp = np.searchsorted(self.cevr, cevr) + 1
        elif isinstance(cevr, tuple):
            ncomp = [np.searchsorted(self.cevr, c) + 1 for c in cevr]

        return ncomp


def svd_wrapper(matrix, mode, ncomp, verbose, full_output=False,
                random_state=None, to_numpy=True):
    """ Wrapper for different SVD libraries (CPU and GPU). 
      
    Parameters
    ----------
    matrix : numpy ndarray, 2d
        2d input matrix.
    mode : {'lapack', 'arpack', 'eigen', 'randsvd', 'cupy', 'eigencupy',
        'randcupy', 'pytorch', 'eigenpytorch', 'randpytorch'}, str optional
        Switch for the SVD method/library to be used.

        ``lapack``: uses the LAPACK linear algebra library through Numpy
        and it is the most conventional way of computing the SVD
        (deterministic result computed on CPU).

        ``arpack``: uses the ARPACK Fortran libraries accessible through
        Scipy (computation on CPU).

        ``eigen``: computes the singular vectors through the
        eigendecomposition of the covariance M.M' (computation on CPU).

        ``randsvd``: uses the randomized_svd algorithm implemented in
        Sklearn (computation on CPU).

        ``cupy``: uses the Cupy library for GPU computation of the SVD as in
        the LAPACK version. `

        `eigencupy``: offers the same method as with the ``eigen`` option
        but on GPU (through Cupy).

        ``randcupy``: is an adaptation of the randomized_svd algorithm,
        where all the computations are done on a GPU (through Cupy). `

        `pytorch``: uses the Pytorch library for GPU computation of the SVD.

        ``eigenpytorch``: offers the same method as with the ``eigen``
        option but on GPU (through Pytorch).

        ``randpytorch``: is an adaptation of the randomized_svd algorithm,
        where all the linear algebra computations are done on a GPU
        (through Pytorch).

    ncomp : int
        Number of singular vectors to be obtained. In the cases when the full
        SVD is computed (LAPACK, ARPACK, EIGEN, CUPY), the matrix of singular 
        vectors is truncated.
    verbose: bool
        If True intermediate information is printed out.
    full_output : bool optional
        If True the 3 terms of the SVD factorization are returned. If ``mode``
        is eigen then only S and V are returned.
    random_state : int, RandomState instance or None, optional
        If int, random_state is the seed used by the random number generator.
        If RandomState instance, random_state is the random number generator.
        If None, the random number generator is the RandomState instance used
        by np.random. Used for ``randsvd`` mode.
    to_numpy : bool, optional
        If True (by default) the arrays computed in GPU are transferred from
        VRAM and converted to numpy ndarrays.

    Returns
    -------
    V : numpy ndarray
        The right singular vectors of the input matrix. If ``full_output`` is
        True it returns the left and right singular vectors and the singular
        values of the input matrix. If ``mode`` is set to eigen then only S and
        V are returned.
    
    References
    ----------
    * For ``lapack`` SVD mode see:
        https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.linalg.svd.html
        http://www.netlib.org/lapack/
    * For ``eigen`` mode see:
        https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.linalg.eigh.html
    * For ``arpack`` SVD mode see:
        https://docs.scipy.org/doc/scipy-0.19.1/reference/generated/scipy.sparse.linalg.svds.html
        http://www.caam.rice.edu/software/ARPACK/
    * For ``randsvd`` SVD mode see:
        https://github.com/scikit-learn/scikit-learn/blob/master/sklearn/utils/extmath.py
        Finding structure with randomness: Stochastic algorithms for constructing
        approximate matrix decompositions
        Halko, et al., 2009 http://arxiv.org/abs/arXiv:0909.4061
    * For ``cupy`` SVD mode see:
        https://docs-cupy.chainer.org/en/stable/reference/generated/cupy.linalg.svd.html
    * For ``eigencupy`` mode see:
        https://docs-cupy.chainer.org/en/master/reference/generated/cupy.linalg.eigh.html
    * For ``pytorch`` SVD mode see:
        http://pytorch.org/docs/master/torch.html#torch.svd
    * For ``eigenpytorch`` mode see:
        http://pytorch.org/docs/master/torch.html#torch.eig

    """
    if matrix.ndim != 2:
        raise TypeError('Input matrix is not a 2d array')

    if ncomp > min(matrix.shape[0], matrix.shape[1]):
        msg = '{} PCs cannot be obtained from a matrix with size [{},{}].'
        msg += ' Increase the size of the patches or request less PCs'
        raise RuntimeError(msg.format(ncomp, matrix.shape[0], matrix.shape[1]))

    if mode == 'eigen':
        # building C as np.dot(matrix.T,matrix) is slower and takes more memory
        C = np.dot(matrix, matrix.T)    # covariance matrix
        e, EV = linalg.eigh(C)          # EVals and EVs
        pc = np.dot(EV.T, matrix)       # PCs using a compact trick when cov is MM'
        V = pc[::-1]                    # reverse since we need the last EVs
        S = np.sqrt(np.abs(e))          # SVals = sqrt(EVals)
        S = S[::-1]                     # reverse since EVals go in increasing order
        for i in range(V.shape[1]):
            V[:, i] /= S    # scaling EVs by the square root of EVals
        V = V[:ncomp]
        if verbose:
            print('Done PCA with numpy linalg eigh functions')

    elif mode == 'lapack':
        # n_frames is usually smaller than n_pixels. In this setting taking
        # the SVD of M' and keeping the left (transposed) SVs is faster than
        # taking the SVD of M (right SVs)
        U, S, V = linalg.svd(matrix.T, full_matrices=False)
        V = V[:ncomp]       # we cut projection matrix according to the # of PCs
        U = U[:, :ncomp]
        S = S[:ncomp]
        if verbose:
            print('Done SVD/PCA with numpy SVD (LAPACK)')

    elif mode == 'arpack':
        U, S, V = svds(matrix, k=ncomp)
        if verbose:
            print('Done SVD/PCA with scipy sparse SVD (ARPACK)')

    elif mode == 'randsvd':
        U, S, V = randomized_svd(matrix, n_components=ncomp, n_iter=2,
                                 transpose='auto', random_state=random_state)
        if verbose:
            print('Done SVD/PCA with randomized SVD')

    elif mode == 'cupy':
        if no_cupy:
            raise RuntimeError('Cupy is not installed')
        a_gpu = cupy.array(matrix)
        a_gpu = cupy.asarray(a_gpu)  # move the data to the current device
        u_gpu, s_gpu, vh_gpu = cupy.linalg.svd(a_gpu, full_matrices=True,
                                               compute_uv=True)
        V = vh_gpu[:ncomp]
        if to_numpy:
            V = cupy.asnumpy(V)
        if full_output:
            S = s_gpu[:ncomp]
            if to_numpy:
                S = cupy.asnumpy(S)
            U = u_gpu[:, :ncomp]
            if to_numpy:
                U = cupy.asnumpy(U)
        if verbose:
            print('Done SVD/PCA with cupy (GPU)')

    elif mode == 'randcupy':
        if no_cupy:
            raise RuntimeError('Cupy is not installed')
        U, S, V = randomized_svd_gpu(matrix, ncomp, n_iter=2, lib='cupy')
        if to_numpy:
            V = cupy.asnumpy(V)
            S = cupy.asnumpy(S)
            U = cupy.asnumpy(U)
        if verbose:
            print('Done randomized SVD/PCA with cupy (GPU)')

    elif mode == 'eigencupy':
        if no_cupy:
            raise RuntimeError('Cupy is not installed')
        a_gpu = cupy.array(matrix)
        a_gpu = cupy.asarray(a_gpu)     # move the data to the current device
        C = cupy.dot(a_gpu, a_gpu.T)    # covariance matrix
        e, EV = cupy.linalg.eigh(C)     # eigenvalues and eigenvectors
        pc = cupy.dot(EV.T, a_gpu)      # using a compact trick when cov is MM'
        V = pc[::-1]                    # reverse to get last eigenvectors
        S = cupy.sqrt(e)[::-1]          # reverse since EVals go in increasing order
        for i in range(V.shape[1]):
            V[:, i] /= S                # scaling by the square root of eigvals
        V = V[:ncomp]
        if to_numpy:
            V = cupy.asnumpy(V)
        if verbose:
            print('Done PCA with cupy eigh function (GPU)')

    elif mode == 'pytorch':
        if no_torch:
            raise RuntimeError('Pytorch is not installed')
        a_gpu = torch.Tensor.cuda(torch.from_numpy(matrix.astype('float32').T))
        u_gpu, s_gpu, vh_gpu = torch.svd(a_gpu)
        V = vh_gpu[:ncomp]
        S = s_gpu[:ncomp]
        U = torch.transpose(u_gpu, 0, 1)[:ncomp]
        if to_numpy:
            V = np.array(V)
            S = np.array(S)
            U = np.array(U)
        if verbose:
            print('Done SVD/PCA with pytorch (GPU)')

    elif mode == 'eigenpytorch':
        if no_torch:
            raise RuntimeError('Pytorch is not installed')
        a_gpu = torch.Tensor.cuda(torch.from_numpy(matrix.astype('float32')))
        C = torch.mm(a_gpu, torch.transpose(a_gpu, 0, 1))
        e, EV = torch.eig(C, eigenvectors=True)
        V = torch.mm(torch.transpose(EV, 0, 1), a_gpu)
        S = torch.sqrt(e[:, 0])
        for i in range(V.shape[1]):
            V[:, i] /= S
        V = V[:ncomp]
        if to_numpy:
            V = np.array(V)
        if verbose:
            print('Done PCA with pytorch eig function')

    elif mode == 'randpytorch':
        if no_torch:
            raise RuntimeError('Pytorch is not installed')
        U, S, V = randomized_svd_gpu(matrix, ncomp, n_iter=2, lib='pytorch')
        if to_numpy:
            V = np.array(V)
            S = np.array(S)
            U = np.array(U)
        if verbose:
            print('Done randomized SVD/PCA with randomized pytorch (GPU)')

    else:
        raise ValueError('The SVD `mode` is not recognized')

    if full_output:
        if mode == 'lapack':
            return V.T, S, U.T
        elif mode == 'pytorch':
            if to_numpy:
                return V.T, S, U.T
            else:
                return torch.transpose(V, 0, 1), S, torch.transpose(U, 0, 1)
        elif mode in ('eigen', 'eigencupy', 'eigenpytorch'):
            return S, V
        else:
            return U, S, V
    else:
        if mode == 'lapack':
            return U.T
        elif mode == 'pytorch':
            return U
        else:
            return V


def get_eigenvectors(ncomp, data, svd_mode, mode='noise', noise_error=1e-3,
                     cevr=0.9, max_evs=None, data_ref=None, debug=False,
                     collapse=False, scaling=None):
    """ Getting ``ncomp`` eigenvectors. Choosing the size of the PCA truncation
    when ``ncomp`` is set to ``auto``. Used in ``pca_annular`` and ``llsg``.
    """
    no_dataref = False
    if data_ref is None:
        no_dataref = True
        data_ref = data

    if max_evs is None:
        max_evs = min(data_ref.shape[0], data_ref.shape[1])

    if ncomp is None:
        raise ValueError('ncomp must be an integer or `auto`')

    if ncomp == 'auto':
        ncomp = 0
        V_big = svd_wrapper(data_ref, svd_mode, max_evs, False)

        if mode == 'noise':
            if not collapse:
                data_ref_sc = matrix_scaling(data_ref, scaling)
                data_sc = matrix_scaling(data, scaling)
            else:
                data_ref_sc = matrix_scaling(data_ref, scaling)
                data_sc = matrix_scaling(data, scaling)

            V_sc = svd_wrapper(data_ref_sc, svd_mode, max_evs, False)

            px_noise = []
            px_noise_decay = 1
            # Noise (px stddev of residuals) to be lower than a given threshold
            while px_noise_decay >= noise_error:
                ncomp += 1
                V = V_sc[:ncomp]
                if no_dataref:
                    transformed = np.dot(data_sc, V.T)
                    reconstructed = np.dot(transformed, V)
                else:
                    transformed = np.dot(V, data_sc)
                    reconstructed = np.dot(transformed.T, V).T
                residuals = data_sc - reconstructed
                if not collapse:
                    curr_noise = np.std(residuals)
                else:
                    curr_noise = np.std((np.median(residuals, axis=0)))
                px_noise.append(curr_noise)
                if ncomp > 1:
                    px_noise_decay = px_noise[-2] - curr_noise
                # print '{} {:.4f} {:.4f}'.format(ncomp, curr_noise, px_noise_decay)
            V = V_big[:ncomp]

        elif mode == 'cevr':
            data_sc = matrix_scaling(data, scaling)
            _, S, _ = svd_wrapper(data_sc, svd_mode, min(data_sc.shape[0],
                                                         data_sc.shape[1]),
                                  False, full_output=True)
            exp_var = (S ** 2) / (S.shape[0] - 1)
            full_var = np.sum(exp_var)
            # % of variance explained by each PC
            explained_variance_ratio = exp_var / full_var
            ratio_cumsum = np.cumsum(explained_variance_ratio)
            ncomp = np.searchsorted(ratio_cumsum, cevr) + 1
            V = V_big[:ncomp]

        if debug:
            print('ncomp', ncomp)

    else:
        # Performing SVD/PCA according to "svd_mode" flag
        ncomp = min(ncomp, min(data_ref.shape[0], data_ref.shape[1]))
        V = svd_wrapper(data_ref, svd_mode, ncomp, verbose=False)

    return V


def randomized_svd_gpu(M, n_components, n_oversamples=10, n_iter='auto',
                       transpose='auto', random_state=0, lib='cupy'):
    """Computes a truncated randomized SVD on GPU. Adapted from Sklearn.

    Parameters
    ----------
    M : ndarray or sparse matrix
        Matrix to decompose
    n_components : int
        Number of singular values and vectors to extract.
    n_oversamples : int (default is 10)
        Additional number of random vectors to sample the range of M so as
        to ensure proper conditioning. The total number of random vectors
        used to find the range of M is n_components + n_oversamples. Smaller
        number can improve speed but can negatively impact the quality of
        approximation of singular vectors and singular values.
    n_iter : int or 'auto' (default is 'auto')
        Number of power iterations. It can be used to deal with very noisy
        problems. When 'auto', it is set to 4, unless `n_components` is small
        (< .1 * min(X.shape)) `n_iter` in which case is set to 7.
        This improves precision with few components.
    transpose : True, False or 'auto' (default)
        Whether the algorithm should be applied to M.T instead of M. The
        result should approximately be the same. The 'auto' mode will
        trigger the transposition if M.shape[1] > M.shape[0] since this
        implementation of randomized SVD tend to be a little faster in that
        case.
    random_state : int, RandomState instance or None, optional (default=None)
        The seed of the pseudo random number generator to use when shuffling
        the data.  If int, random_state is the seed used by the random number
        generator; If RandomState instance, random_state is the random number
        generator; If None, the random number generator is the RandomState
        instance used by `np.random`.
    lib : {'cupy', 'pytorch'}, str optional
        Chooses the GPU library to be used.

    Notes
    -----
    This algorithm finds a (usually very good) approximate truncated
    singular value decomposition using randomization to speed up the
    computations. It is particularly fast on large matrices on which
    you wish to extract only a small number of components. In order to
    obtain further speed up, `n_iter` can be set <=2 (at the cost of
    loss of precision).

    References
    ----------
    * Finding structure with randomness: Stochastic algorithms for constructing
      approximate matrix decompositions
      Halko, et al., 2009 http://arxiv.org/abs/arXiv:0909.4061
    * A randomized algorithm for the decomposition of matrices
      Per-Gunnar Martinsson, Vladimir Rokhlin and Mark Tygert
    * An implementation of a randomized algorithm for principal component
      analysis
      A. Szlam et al. 2014
    """
    random_state = check_random_state(random_state)
    n_random = n_components + n_oversamples
    n_samples, n_features = M.shape

    if n_iter == 'auto':
        # Checks if the number of iterations is explicitly specified
        n_iter = 7 if n_components < .1 * min(M.shape) else 4

    if transpose == 'auto':
        transpose = n_samples < n_features
    if transpose:
        M = M.T # this implementation is a bit faster with smaller shape[1]

    if lib == 'cupy':
        M = cupy.array(M)
        M = cupy.asarray(M)

        # Generating normal random vectors with shape: (M.shape[1], n_random)
        Q = random_state.normal(size=(M.shape[1], n_random))
        Q = cupy.array(Q)
        Q = cupy.asarray(Q)

        # Perform power iterations with Q to further 'imprint' the top
        # singular vectors of M in Q
        for i in range(n_iter):
            Q = cupy.dot(M, Q)
            Q = cupy.dot(M.T, Q)

        # Sample the range of M by linear projection of Q.
        # Extract an orthonormal basis
        Q, _ = cupy.linalg.qr(cupy.dot(M, Q), mode='reduced')

        # project M to the (k + p) dimensional space using the basis vectors
        B = cupy.dot(Q.T, M)

        B = cupy.array(B)
        Q = cupy.array(Q)
        # compute the SVD on the thin matrix: (k + p) wide
        Uhat, s, V = cupy.linalg.svd(B, full_matrices=False, compute_uv=True)
        del B
        U = cupy.dot(Q, Uhat)

        if transpose:
            # transpose back the results according to the input convention
            return (V[:n_components, :].T, s[:n_components],
                    U[:,:n_components].T)
        else:
            return U[:, :n_components], s[:n_components], V[:n_components, :]

    elif lib == 'pytorch':
        M_gpu = torch.Tensor.cuda(torch.from_numpy(M.astype('float32')))

        # Generating normal random vectors with shape: (M.shape[1], n_random)
        Q = torch.cuda.FloatTensor(M_gpu.shape[1], n_random).normal_()

        # Perform power iterations with Q to further 'imprint' the top
        # singular vectors of M in Q
        for i in range(n_iter):
            Q = torch.mm(M_gpu, Q)
            Q = torch.mm(torch.transpose(M_gpu, 0, 1), Q)

        # Sample the range of M by linear projection of Q.
        # Extract an orthonormal basis
        Q, _ = torch.qr(torch.mm(M_gpu, Q))

        # project M to the (k + p) dimensional space using the basis vectors
        B = torch.mm(torch.transpose(Q, 0, 1), M_gpu)

        # compute the SVD on the thin matrix: (k + p) wide
        Uhat, s, V = torch.svd(B)
        del B
        U = torch.mm(Q, Uhat)

        if transpose:
            # transpose back the results according to the input convention
            return (torch.transpose(V[:n_components, :], 0, 1),
                    s[:n_components],
                    torch.transpose(U[:, :n_components], 0, 1))
        else:
            return U[:, :n_components], s[:n_components], V[:n_components, :]





"""
subpixel shifting.
"""


__author__ = 'Ralf Farkas'
__all__ = []


import numpy as np


def calc_psf_shift_subpix(psf, precision):
    """
    Compute a stack of subpixel-shifted versions of the PSF.

    Parameters
    ----------
    psf : 2d numpy ndarray
        The PSF that is to be shifted. Assumed square.
    precision : int
        Number of pixel subdivisions for the planet's signal pattern
        computation, i.e., inverse of the shifting pitch


    Returns
    -------
    psf_cube : 4d ndarray
        4d array that contains all the shifted versions of the PSF. The first
        two indices contain the fraction of the shift in the x and y directions,
        the last two refer to the spatial position on the grid. The shape
        of the array is (precision+1, precision+1, n, n), where n is the
        side of the PSF used. The image ``psf_cube[i, j, :, :]`` thus
        corresponds to ``psf`` shifted by ``i/precision`` in x and
        ``j/precision`` in y.

    Notes
    -----
    The IDL implementation has an inversed shape of
    (n, n, precision+1, precision+1), and indexing works like
    ``psf_cube[*, *, i, j]``, where ``i`` is the column and ``j`` the row.

    based on ``LibAndromeda/oneralib/calc_psf_shift_subpix.pro``,
    v1.2 2010/05/27

    The IDL version has an optional keyword ``INTERP_INPUT``, which is not used
    in andromeda. It was not implemented.

    """
    n = psf.shape[0]
    psf_cube = np.zeros((precision+1, precision+1, n, n))

    for i_column in range(precision+1):
        decalx = i_column/precision
        for j_row in range(precision+1):
            decaly = j_row/precision
            psf_cube[j_row, i_column] = subpixel_shift(psf, decalx, decaly)

    return psf_cube


def subpixel_shift(image, xshift, yshift):
    """
    Subpixel shifting of ``image`` using fourier transformation.

    Parameters
    ----------
    image : 2d numpy ndarray
        The image to be shifted.
    xshift : float
        Amount of desired shift in X direction.
    yshift : float
        Amount of desired shift in Y direction.

    Returns
    -------
    shifted_image : 2d ndarray
        Input ``image`` shifted by ``xshift`` and ``yshift``.

    Notes
    -----
    based on ``LibAndromeda/oneralib/subpixel_shift.pro``, v1.3 2009/05/28

    """
    npix = image.shape[0]

    if npix != image.shape[1]:
        raise ValueError("`image` must be square")

    ramp = np.outer(np.ones(npix), np.arange(npix) - npix/2)
    tilt = (-2*np.pi / npix) * (xshift*ramp + yshift*ramp.T)
    fact = np.fft.fftshift(np.cos(tilt) + 1j*np.sin(tilt))

    image_ft = np.fft.fft2(image)  # no np.fft.fftshift applied!
    shifted_image = np.fft.ifft2(image_ft * fact).real

    return shifted_image



#! /usr/bin/env python

"""
Module for simplex or grid search of best fit spectrum in a template library.
TBD: implement multiprocessing version.
"""

__author__ = 'V. Christiaens'
__all__ = ['best_fit_tmp',
           'get_chi']

from datetime import datetime
import numpy as np
import os
from scipy.optimize import minimize
from ..conf import time_ini, timing, time_fin
from .chi import gof_scal
from .model_resampling import resample_model
from .utils_spec import extinction, find_nearest
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)


def get_chi(lbda_obs, spec_obs, err_obs, tmp_name, tmp_reader, 
            search_mode='simplex', lambda_scal=None, scale_range=(0.1,10,0.01), 
            ext_range=None, dlbda_obs=None, instru_corr=None, instru_fwhm=None, 
            instru_idx=None, filter_reader=None, simplex_options=None,
            red_chi2=True, remove_nan=False, force_continue=False, 
            verbose=False, **kwargs):
    """ Routine calculating chi^2, optimal scaling factor and optimal 
    extinction for a given template spectrum to match an observed spectrum.
    
    Parameters
    ----------
    lbda_obs : numpy 1d ndarray or list
        Wavelength of observed spectrum. If several instruments, should be 
        ordered per instrument, not necessarily as monotonically increasing 
        wavelength. Hereafter, n_ch = len(lbda_obs).
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of lbda_obs.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. If 2d array, should be [2,n_ch]
        where the first (resp. second) column corresponds to lower (upper) 
        uncertainty, and n_ch is the length of lbda_obs and spec_obs.
    tmp_name :  str
        Template spectrum filename.
    tmp_reader : python routine
        External routine that reads a model file and returns a 3D numpy array, 
        where the first column corresponds to wavelengths, the second 
        contains flux values, and the third the uncertainties on the flux.
    search_mode: str, opt {'simplex', 'grid'}
        How is the best fit template found? Simplex or grid search.
    lambda_scal: float, optional
        Wavelength where a first scaling will be performed between template
        and observed spectra. If not provided, the middle wavelength of the 
        osberved spectra will be considered.
    scale_range: tuple, opt
        If grid search, this parameter should be provided as a tuple of 3 
        floats: lower limit, upper limit and step of the grid search for the 
        scaling factor to be applied AFTER the first rough scaling (i.e.
        scale_range should always encompass 1).
    ext_range: tuple or None, opt
        If None: differential extinction is not to be considered as a free 
        parameter. Elif a tuple of 3 floats is provided, differential extinction 
        will be considered, with the floats as lower limit, upper limit and step 
        of the grid search.
        Note: if simplex search, the range is still used to set a chi of 
        np.inf outside of the range.
    dlbda_obs: numpy 1d ndarray or list, optional
        Spectral channel width for the observed spectrum. It should be provided 
        IF one wants to weigh each point based on the spectral 
        resolution of the respective instruments (as in Olofsson et al. 2016).
    instru_corr : numpy 2d ndarray or list, optional
        Spectral correlation throughout post-processed images in which the 
        spectrum is measured. It is specific to the combination of instrument, 
        algorithm and radial separation of the companion from the central star.
        Can be computed using distances.spectral_correlation(). In case of
        a spectrum obtained with different instruments, build it with
        distances.combine_corrs(). If not provided, it will consider the 
        uncertainties in each spectral channels are independent. See Greco & 
        Brandt (2017) for details.
    instru_fwhm : float or list, optional
        The instrumental spectral fwhm provided in nm. This is used to convolve
        the model spectrum. If several instruments are used, provide a list of 
        instru_fwhm values, one for each instrument whose spectral resolution
        is coarser than the model - including broad band
        filter FWHM if relevant.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to n_instru. Zero for points 
        that don't correspond to any instru_fwhm provided above, and i in 
        [1,n_instru] for points associated to instru_fwhm[i-1]. This parameter 
        must be provided if the spectrum consists of points obtained with 
        different instruments.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_fwhm, the default format assumed for the files:
        - first row containing header
        - starting from 2nd row: 1st column: WL in mu, 2nd column: transmission
        Note: files should all have the same format and wavelength units.
    red_chi2: bool, optional
        Whether to compute the reduced chi square. If False, considers chi^2.
    remove_nan: bool, optional
        Whether to remove NaN values from template spectrum BEFORE resampling
        to the wavelength sampling of the observed spectrum. Whether it is set
        to True or False, a check is made just before chi^2 is calculated 
        (after resampling), and only non-NaN values will be considered.
    simplex_options: dict, optional
        The scipy.optimize.minimize simplex (Nelder-Mead) options.
    force_continue: bool, optional
        In case of issue with the fit, whether to continue regardless (this may
        be useful in an uneven spectral library, where some templates have too
        few points for the fit to be performed).
    verbose: str, optional
        Whether to print more information when fit fails.
    **kwargs: optional
        Optional arguments to the scipy.optimize.minimize function.
        
    Returns
    -------
    best_chi: float
        goodness of fit scored by the template
    best_scal:
        best-fit scaling factor for the considered template
    best_ext:
        best-fit optical extinction for the considered template
        
    """
    # read template spectrum
    try:
        lbda_tmp, spec_tmp, spec_tmp_err = tmp_reader(tmp_name, 
                                                  verbose=bool(verbose))
    except:
        msg = "{} could not be opened. Corrupt file?".format(tmp_name)
        if force_continue:
            if verbose:
                print(msg)
            return np.inf, 1, 0, 1
        else:
            raise ValueError(msg)
            
    # look for any nan and replace
    if remove_nan:
        if np.isnan(spec_tmp).any() or np.isnan(spec_tmp_err).any():
            bad_idx1 = np.where(np.isnan(spec_tmp))[0]
            bad_idx2 = np.where(np.isnan(spec_tmp_err))[1]
            all_bad = np.concatenate(bad_idx1,bad_idx2)
            nch = len(lbda_tmp)
            new_lbda = [lbda_tmp[i] for i in range(nch) if i not in all_bad]
            new_spec = [spec_tmp[i] for i in range(nch) if i not in all_bad]
            new_err = [spec_tmp_err[i] for i in range(nch) if i not in all_bad]
            lbda_tmp = np.array(new_lbda)
            spec_tmp = np.array(new_spec)
            spec_tmp_err = np.array(new_err)
        
    # don't consider template spectra whose range is smaller than observed one
    if lbda_obs[0] < lbda_tmp[0] or lbda_obs[-1] > lbda_tmp[-1]:
        msg = "Wavelength range of template {} ({:.2f}, {:.2f})mu too short"
        msg+= " compared to that of observed spectrum ({:.2f}, {:.2f})mu"
        if force_continue:
            if verbose:
                print(msg.format(tmp_name, lbda_tmp[0],lbda_tmp[-1],
                                        lbda_obs[0],lbda_obs[-1]))
            return np.inf, 1, 0, len(lbda_tmp)-2
        else:
            raise ValueError(msg.format(tmp_name, lbda_tmp[0],lbda_tmp[-1],
                                        lbda_obs[0],lbda_obs[-1]))
    
    # resample as observed spectrum
    try:
        _, spec_tmp = resample_model(lbda_obs, lbda_tmp, spec_tmp, 
                                     dlbda_obs=dlbda_obs, 
                                     instru_fwhm=instru_fwhm, 
                                     instru_idx=instru_idx, 
                                     filter_reader=filter_reader)
        lbda_tmp, spec_tmp_err = resample_model(lbda_obs, lbda_tmp, 
                                                spec_tmp_err, 
                                                dlbda_obs=dlbda_obs, 
                                                instru_fwhm=instru_fwhm, 
                                                instru_idx=instru_idx, 
                                                filter_reader=filter_reader)
    except:
        msg = "Issue with resampling of template {}. Does the wavelength "
        msg+= "range extend far enough ({:.2f}, {:.2f})mu?"
        if force_continue:
            if verbose:
                print(msg.format(tmp_name, lbda_tmp[0],lbda_tmp[-1]))
            return np.inf, 1, 0, len(lbda_tmp)-2
        else:
            raise ValueError(msg.format(tmp_name, lbda_tmp[0],lbda_tmp[-1]))
    
    # first rescaling fac
    if not lambda_scal:
        lambda_scal = (lbda_obs[0]+lbda_obs[-1])/2
    idx_cen = find_nearest(lbda_obs, lambda_scal)
    scal_fac = spec_obs[idx_cen]/spec_tmp[idx_cen]
    spec_tmp*=scal_fac
    spec_tmp_err*=scal_fac
    
    # combine observed and template uncertainties
    # EDIT: Don't: the  best fit will be the most noisy tmp of the library!)
    #err_obs = np.sqrt(np.power(spec_tmp_err,2)+np.power(err_obs,2))
    
    
    # only consider non-zero and non-nan values for chi^2 calculation
    cond1 = np.where(np.isfinite(spec_tmp))[0]
    cond2 = np.where(np.isfinite(err_obs))[0] 
    cond3 = np.where(spec_tmp>0)[0]
    all_conds = np.sort(np.unique(np.concatenate((cond1,cond2,cond3))))
    ngood_ch = len(all_conds)
    good_ch = (all_conds,)
    lbda_obs = lbda_obs[good_ch]
    spec_obs = spec_obs[good_ch]
    err_obs = err_obs[good_ch]
    spec_tmp = spec_tmp[good_ch]
    
    n_dof = ngood_ch-1-(ext_range is not None)
    if n_dof <= 0:
        msg = "Not enough dof with remaining points of template spectrum {}"
        if force_continue:
            if verbose:
                print(msg.format(tmp_name))
            return np.inf, 1, 0, n_dof
        else:
            raise ValueError(msg.format(tmp_name))
    
    # simplex search
    if search_mode == 'simplex':
        if simplex_options is None:
            simplex_options = {'xatol': 1e-6, 'fatol': 1e-6, 'maxiter': 1000,
                               'maxfev': 5000}
        if not ext_range:
            p = (1,)
        else:
            AV_ini = (ext_range[0]+ext_range[1])/2
            p = (1,AV_ini)
        try:
            res = minimize(gof_scal, p, args=(lbda_obs, spec_obs, err_obs, 
                                              lbda_obs, spec_tmp, dlbda_obs, 
                                              instru_corr, instru_fwhm, 
                                              instru_idx, filter_reader,
                                              ext_range),
                           method='Nelder-Mead', options=simplex_options, 
                           **kwargs)
        except:
            msg = "Issue with simplex minimization for template {}. "
            msg+= "Try grid search?"
            if force_continue:
                if verbose:
                    print(msg.format(tmp_name))
                return np.inf, 1, 0, n_dof
            else:
                raise ValueError(msg.format(tmp_name))
        best_chi = res.fun
        if not ext_range:
            best_scal = res.x
            best_ext = 0
        else:
            best_scal, best_ext = res.x
        
    # or grid search        
    elif search_mode == 'grid':
        test_scale = np.arange(scale_range[0], scale_range[1], scale_range[2])
        n_test = len(test_scale)
        if ext_range is None:
            test_ext = np.array([0])
            n_ext = 1
        elif isinstance(ext_range, tuple) and len(ext_range)==3:
            test_ext = np.arange(ext_range[0], ext_range[1], ext_range[2])
            n_ext = len(test_ext)
        else:
            raise TypeError("ext_range can only be None or tuple of length 3")

        chi = np.zeros([n_test,n_ext])
        
        for cc, scal in enumerate(test_scale):
            for ee, AV in enumerate(test_ext):
                p = (scal,AV)
                chi[cc,ee] = gof_scal(p, lbda_obs, spec_obs, err_obs, lbda_obs, 
                                      spec_tmp, dlbda_obs=dlbda_obs, 
                                      instru_corr=instru_corr, 
                                      instru_fwhm=instru_fwhm, 
                                      instru_idx=instru_idx, 
                                      filter_reader=filter_reader,
                                      ext_range=ext_range) 
        best_chi = np.nanmin(chi)
        best_idx = np.nanargmin(chi)
        best_idx = np.unravel_index(best_idx,chi.shape)
        best_scal = test_scale[best_idx[0]]
        best_ext = test_ext[best_idx[1]]
    
    else:
        msg = "Search mode not recognised. Should be 'simplex' or 'grid'."
        raise TypeError(msg)
    
    if red_chi2:
        best_chi /= n_dof
        
    
    return best_chi, best_scal*scal_fac, best_ext, n_dof



def best_fit_tmp(lbda_obs, spec_obs, err_obs, tmp_reader, search_mode='simplex',
                 n_best=1, lambda_scal=None, scale_range=(0.1,10,0.01), 
                 ext_range=None, simplex_options=None, dlbda_obs=None, 
                 instru_corr=None, instru_fwhm=None, instru_idx=None, 
                 filter_reader=None, lib_dir='tmp_lib/', tmp_endswith='.fits', 
                 red_chi2=True, remove_nan=False, nproc=1, verbosity=0, 
                 force_continue=False, **kwargs):
    """ Finds the best fit template spectrum to a given observed spectrum, 
    within a spectral library.  By default, a single free parameter is 
    considered: the scaling factor of the spectrum. A first automatic scaling 
    is performed by comparing the flux of the observed and template spectra at 
    lambda_scal. Then a more refined scaling is performed, either through 
    simplex or grid search (within scale_range).
    If fit_extinction is set to True, the exctinction is also considered as a 
    free parameter.
    
    Parameters
    ----------
    lbda_obs : numpy 1d ndarray or list
        Wavelength of observed spectrum. If several instruments, should be 
        ordered per instrument, not necessarily as monotonically increasing 
        wavelength. Hereafter, n_ch = len(lbda_obs).
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of lbda_obs.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. If 2d array, should be [2,n_ch]
        where the first (resp. second) column corresponds to lower (upper) 
        uncertainty, and n_ch is the length of lbda_obs and spec_obs.
    tmp_reader : python routine
        External routine that reads a model file and returns a 3D numpy array, 
        where the first column corresponds to wavelengths, the second 
        contains flux values, and the third the uncertainties on the flux.
    search_mode: str, optional, {'simplex','grid'}
        How is the best fit template found? Simplex or grid search.
    n_best: int, optional
        Number of best templates to be returned (default: 1)
    lambda_scal: float, optional
        Wavelength where a first scaling will be performed between template
        and observed spectra. If not provided, the middle wavelength of the 
        osberved spectra will be considered.
    scale_range: tuple, opt
        If grid search, this parameter should be provided as a tuple of 3 
        floats: lower limit, upper limit and step of the grid search for the 
        scaling factor to be applied AFTER the first rough scaling (i.e.
        scale_range should always encompass 1).
    ext_range: tuple or None, opt
        If None: differential extinction is not to be considered as a free 
        parameter. Elif a tuple of 3 floats is provided, differential extinction 
        will be considered, with the floats as lower limit, upper limit and step 
        of the grid search.
        Note: if simplex search, the range is still used to set a chi of 
        np.inf outside of the range.
    dlbda_obs: numpy 1d ndarray or list, optional
        Spectral channel width for the observed spectrum. It should be provided 
        IF one wants to weigh each point based on the spectral 
        resolution of the respective instruments (as in Olofsson et al. 2016).
    instru_corr : numpy 2d ndarray or list, optional
        Spectral correlation throughout post-processed images in which the 
        spectrum is measured. It is specific to the combination of instrument, 
        algorithm and radial separation of the companion from the central star.
        Can be computed using distances.spectral_correlation(). In case of
        a spectrum obtained with different instruments, build it with
        distances.combine_corrs(). If not provided, it will consider the 
        uncertainties in each spectral channels are independent. See Greco & 
        Brandt (2017) for details.
    instru_fwhm : float OR list of either floats or strings, optional
        The instrumental spectral fwhm provided in nm. This is used to convolve
        the model spectrum. If several instruments are used, provide a list of 
        instru_fwhm values, one for each instrument whose spectral resolution
        is coarser than the model - including broad band filter FWHM if 
        relevant.
        If strings are provided, they should correspond to filenames (including 
        full paths) of text files containing the filter information for each 
        observed wavelength. Strict format: 
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to n_instru. Zero for points 
        that don't correspond to any instru_fwhm provided above, and i in 
        [1,n_instru] for points associated to instru_fwhm[i-1]. This parameter 
        must be provided if the spectrum consists of points obtained with 
        different instruments.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_fwhm, the default file reader will be used. 
        It assumes the following format for the files:
        - first row containing header
        - starting from 2nd row: 1st column: wavelength, 2nd col.: transmission
        - Unit of wavelength can be provided in parentheses of first header key
        name: e.g. "WL(AA)" for angstrom, "wavelength(mu)" for micrometer or
        "lambda(nm)" for nanometer. Note: Only what is in parentheses matters.
        Important: filter files should all have the same format and WL units.
    simplex_options: dict, optional
        The scipy.optimize.minimize simplex (Nelder-Mead) options.
    red_chi2: bool, optional
        Whether to compute the reduced chi square. If False, considers chi^2.
    remove_nan: bool, optional
        Whether to remove NaN values from template spectrum BEFORE resampling
        to the wavelength sampling of the observed spectrum. Whether it is set
        to True or False, a check is made just before chi^2 is calculated 
        (after resampling), and only non-NaN values will be considered.
    nproc: int, optional
        The number of processes to use for parallelization.
    verbosity: 0, 1 or 2, optional
        Verbosity level. 0 for no output and 2 for full information.
    force_continue: bool, optional
        In case of issue with the fit, whether to continue regardless (this may
        be useful in an uneven spectral library, where some templates have too
        few points for the fit to be performed).
    **kwargs: optional
        Optional arguments to the scipy.optimize.minimize function
         
    Returns
    -------
    final_tmpname: tuple of n_best str
        Best-fit template filenames
    final_tmp: tuple of n_best 3D numpy array
        Best-fit template spectra (3D: lbda+spec+spec_err)
    final_chi: 1D numpy array of length n_best
        Best-fit template chi^2
    final_params: 2D numpy array (2xn_best)
        Best-fit parameters (optimal scaling and optical extinction). Note if 
        extinction is not fitted, optimal AV will be set to 0.
        
    """
    # create list of template filenames
    tmp_filelist = [x for x in os.listdir(lib_dir) if x.endswith(tmp_endswith)]
    n_tmp = len(tmp_filelist)
    
    if verbosity > 0:
        start_time = time_ini()
        int_time = time_ini()
        print("{:.0f} template spectra will be tested. \n".format(n_tmp))
        print("****************************************\n")
    
    chi = np.ones(n_tmp)
    scal = np.ones_like(chi)
    ext = np.zeros_like(chi)
    n_dof =  np.ones_like(chi)
    counter = 0
    
    if nproc == 1:
        for tt in range(n_tmp):
            if verbosity>0 and search_mode=='simplex':
                print('Nelder-Mead minimization is running...')
            chi[tt], scal[tt], ext[tt], n_dof[tt] = get_chi(lbda_obs, spec_obs, err_obs, 
                                                 tmp_filelist[tt], tmp_reader, 
                                                 search_mode=search_mode, 
                                                 scale_range=scale_range, 
                                                 ext_range=ext_range,
                                                 lambda_scal=lambda_scal,
                                                 dlbda_obs=dlbda_obs, 
                                                 instru_corr=instru_corr, 
                                                 instru_fwhm=instru_fwhm, 
                                                 instru_idx=instru_idx, 
                                                 filter_reader=filter_reader,
                                                 simplex_options=simplex_options,
                                                 red_chi2=red_chi2,
                                                 remove_nan=remove_nan,
                                                 force_continue=force_continue,
                                                 verbose=verbosity,
                                                 **kwargs)
            
            if chi[tt]<np.inf:
                counter+=1
            elif verbosity>0:
                msg = "{:.0f}/{:.0f} ({}) FAILED"
                if np.isnan(chi[tt]):
                    msg += " (simplex did not converge)"
                print(msg.format(tt, n_tmp, tmp_filelist[tt]))

            if verbosity > 0 and tt==0:
                msg = "{:.0f}/{:.0f}: done in {}s"
                indiv_time = time_fin(start_time)
                print(msg.format(tt, n_tmp, indiv_time))
                now = datetime.now()
                delta_t = now.timestamp()-start_time.timestamp()
                tot_time = n_tmp*delta_t/60
                print("Fit may take a total of ~{:.0f}min \n".format(tot_time))
                int_time = time_ini(verbose=False)
            elif verbosity > 1:
                msg = "{:.0f}/{:.0f}: done in {}s \n"
                indiv_time = time_fin(int_time)
                int_time = time_ini(verbose=False)
                print(msg.format(tt, n_tmp, indiv_time))
    else:
        raise ValueError("multiprocessing mode yet to be implemented")
        
    if verbosity > 0:
        print("****************************************\n")
        msg = "{:.0f}/{:.0f} template spectra were fitted. \n"
        print(msg.format(counter, n_tmp))
        timing(start_time)
        
    return best_n_tmp(chi, scal, ext, n_dof, tmp_filelist, tmp_reader, n_best=n_best,
                      verbose=True)
    
    
def best_n_tmp(chi, scal, ext, n_dof, tmp_filelist, tmp_reader, n_best=1, 
               verbose=False):
    """
    Routine returning the n best template spectra.
    
    Parameters
    ----------
    lbda_obs : numpy 1d ndarray or list
        Wavelength of observed spectrum. If several instruments, should be 
        ordered per instrument, not necessarily as monotonically increasing 
        wavelength. Hereafter, n_ch = len(lbda_obs).

    Returns
    -------
    final_tmpname: tuple of n_best str
        Best-fit template filenames
    final_tmp: tuple of n_best 3D numpy array
        Best-fit template spectra (3D: lbda+spec+spec_err)
    final_chi: 1D numpy array of length n_best
        Best-fit template chi^2
    final_params: 2D numpy array (2xn_best)
        Best-fit parameters (optimal scaling and optical extinction). Note if 
        extinction is not fitted, optimal AV will be set to 0.
        
    """
    # sort from best to worst match
    order = np.argsort(chi)
    sort_chi = chi[order]
    sort_scal = scal[order]
    sort_ext = ext[order]
    sort_ndof = n_dof[order]
    sort_filelist = [tmp_filelist[i] for i in order]
    
    if verbose:
        print("best chi: ", sort_chi[:n_best])
        print("best scale fac: ", sort_scal[:n_best])
        print("n_dof: ", sort_ndof[:n_best])
    # take the n_best first ones
    best_tmp = []
    for n in range(n_best):
        lbda_tmp, spec_tmp, spec_tmp_err = tmp_reader(sort_filelist[n])
        Albdas = extinction(lbda_tmp,abs(sort_ext[n]))
        extinc_fac = np.power(10.,-Albdas/2.5)
        if sort_ext[n]>0:
            final_scal = sort_scal[n]*extinc_fac
        elif sort_ext[n]<0:
            final_scal = sort_scal[n]/extinc_fac
        else:
            final_scal = sort_scal[n]
        best_tmp.append(np.array([lbda_tmp, spec_tmp*final_scal, 
                                  spec_tmp_err*final_scal]))
        if verbose:
            msg = "The best template #{:.0f} is: {} "
            msg+="(Delta A_V={:.1f}mag)\n"
            print(msg.format(n, sort_filelist[n], sort_ext[n]))
            
    best_tmpname = tuple(sort_filelist[:n_best])
    best_tmp = tuple(best_tmp)
    best_params = np.array([sort_scal[:n_best],sort_ext[:n_best]])
    
    return best_tmpname, best_tmp, sort_chi[:n_best], best_params, sort_ndof[:n_best]


#! /usr/bin/env python

"""
Module with S/N calculation functions.
We strongly recommend users to read Mawet et al. (2014) before using routines 
of this module: https://ui.adsabs.harvard.edu/abs/2014ApJ...792...97M/abstract
"""

__author__ = 'Carlos Alberto Gomez Gonzalez, O. Absil @ ULg, V. Christiaens'
__all__ = ['snr',
           'snrmap',
           'significance',
           'frame_report']

import numpy as np
import photutils
from scipy.stats import norm, t
from hciplot import plot_frames
from skimage.draw import disk, circle_perimeter
from matplotlib import pyplot as plt
from astropy.convolution import convolve, Tophat2DKernel
from astropy.stats import median_absolute_deviation as mad
from multiprocessing import cpu_count
from ..conf.utils_conf import pool_map, iterable, sep
from ..conf import time_ini, timing, check_array
from ..var import get_annulus_segments, frame_center, dist


def snrmap(array, fwhm, approximated=False, plot=False, known_sources=None,
           nproc=None, array2=None, use2alone=False, 
           exclude_negative_lobes=False, verbose=True, **kwargs):
    """Parallel implementation of the S/N map generation function. Applies the
    S/N function (small samples penalty) at each pixel.
    
    The S/N is computed as in Mawet et al. (2014) for each radial separation.    
    https://ui.adsabs.harvard.edu/abs/2014ApJ...792...97M/abstract
    
    *** DISCLAIMER ***
    Signal-to-noise ratio is not significance! For a conversion from snr to 
    n-sigma (i.e. the equivalent confidence level of a Gaussian n-sigma), use 
    the significance() function.    
    
    
    Parameters
    ----------
    array : numpy ndarray
        Input frame (2d array).
    fwhm : float
        Size in pixels of the FWHM.
    approximated : bool, optional
        If True, an approximated S/N map is generated.
    plot : bool, optional
        If True plots the S/N map. False by default.
    known_sources : None, tuple or tuple of tuples, optional
        To take into account existing sources. It should be a tuple of float/int
        or a tuple of tuples (of float/int) with the coordinate(s) of the known
        sources.
    nproc : int or None
        Number of processes for parallel computing.
    array2 : numpy ndarray, optional
        Additional image (e.g. processed image with negative derotation angles) 
        enabling to have more noise samples. Should have the 
        same dimensions as array.
    use2alone: bool, optional
        Whether to use array2 alone to estimate the noise (might be useful to 
        estimate the snr of extended disk features).
    verbose: bool, optional
        Whether to print timing or not.
    **kwargs : dictionary, optional
        Arguments to be passed to ``plot_frames`` to customize the plot (and to
        save it to disk).

    Returns
    -------
    snrmap : 2d numpy ndarray
        Frame with the same size as the input frame with each pixel.
    """
    if verbose:
        start_time = time_ini()

    check_array(array, dim=2, msg='array')
    sizey, sizex = array.shape
    snrmap_array = np.zeros_like(array)
    width = min(sizey, sizex) / 2 - 1.5 * fwhm
    mask = get_annulus_segments(array, (fwhm / 2) + 2, width, mode="mask")[0]
    mask = np.ma.make_mask(mask)
    # by making a bool mask *after* applying the mask to the array, we also mask
    # out zero values from the array. This logic cannot be simplified by using
    # mode="ind"!
    yy, xx = np.where(mask)
    coords = zip(xx, yy)

    if nproc is None:
        nproc = cpu_count() // 2        # Hyper-threading doubles the # of cores

    if known_sources is None:

        # proxy to S/N calculation
        if approximated:
            cy, cx = frame_center(array)
            tophat_kernel = Tophat2DKernel(fwhm / 2)
            array = convolve(array, tophat_kernel)
            width = min(sizey, sizex) / 2 - 1.5 * fwhm
            mask = get_annulus_segments(array, (fwhm / 2) + 1, width - 1,
                                        mode="mask")[0]
            mask = np.ma.make_mask(mask)
            yy, xx = np.where(mask)
            coords = [(int(x), int(y)) for (x, y) in zip(xx, yy)]
            res = pool_map(nproc, _snr_approx, array, iterable(coords), fwhm,
                           cy, cx)
            res = np.array(res)
            yy = res[:, 0]
            xx = res[:, 1]
            snr_value = res[:, 2]
            snrmap_array[yy.astype(int), xx.astype(int)] = snr_value

        # computing s/n map with Mawet+14 definition
        else:
            res = pool_map(nproc, snr, array, iterable(coords), fwhm, True,
                           array2, use2alone, exclude_negative_lobes)
            res = np.array(res)
            yy = res[:, 0]
            xx = res[:, 1]
            snr_value = res[:, -1]
            snrmap_array[yy.astype('int'), xx.astype('int')] = snr_value

    # masking known sources
    else:
        if not isinstance(known_sources, tuple):
            raise TypeError("`known_sources` must be a tuple or tuple of "
                            "tuples")
        else:
            source_mask = np.zeros_like(array)
            if isinstance(known_sources[0], tuple):
                for coor in known_sources:
                    source_mask[coor[::-1]] = 1
            elif isinstance(known_sources[0], int):
                source_mask[known_sources[1], known_sources[0]] = 1
            else:
                raise TypeError("`known_sources` seems to have wrong type. It "
                                "must be a tuple of ints or tuple of tuples "
                                "(of ints)")

        # checking the mask with the sources
        if source_mask[source_mask == 1].shape[0] > 50:
            msg = 'Input source mask is too crowded (check its validity)'
            raise RuntimeError(msg)

        soury, sourx = np.where(source_mask == 1)
        sources = []
        coor_ann = []
        arr_masked_sources = array.copy()
        centery, centerx = frame_center(array)
        for y, x in zip(soury, sourx):
            radd = dist(centery, centerx, int(y), int(x))
            if int(radd) < centery - np.ceil(fwhm):
                sources.append((y, x))

        for source in sources:
            y, x = source
            radd = dist(centery, centerx, int(y), int(x))
            anny, annx = get_annulus_segments(array, int(radd-fwhm),
                                              int(np.round(3 * fwhm)))[0]

            ciry, cirx = disk((y, x), int(np.ceil(fwhm)))
            # masking the sources positions (using the MAD of pixels in annulus)
            arr_masked_sources[ciry, cirx] = mad(array[anny, annx])

            # S/Ns of annulus without the sources
            coor_ann = [(x, y) for (x, y) in zip(annx, anny) if (x, y) not in
                        zip(cirx, ciry)]
            res = pool_map(nproc, snr, arr_masked_sources, iterable(coor_ann),
                           fwhm, True, array2, use2alone, 
                           exclude_negative_lobes)
            res = np.array(res)
            yy_res = res[:, 0]
            xx_res = res[:, 1]
            snr_value = res[:, 4]
            snrmap_array[yy_res.astype('int'), xx_res.astype('int')] = snr_value
            coor_ann += coor_ann

        # S/Ns of the rest of the frame without the annulus
        coor_rest = [(x, y) for (x, y) in zip(xx, yy) if (x, y) not in coor_ann]
        res = pool_map(nproc, snr, array, iterable(coor_rest), fwhm, True,
                       array2, use2alone, exclude_negative_lobes)
        res = np.array(res)
        yy_res = res[:, 0]
        xx_res = res[:, 1]
        snr_value = res[:, 4]
        snrmap_array[yy_res.astype('int'), xx_res.astype('int')] = snr_value

    if plot:
        plot_frames(snrmap_array, colorbar=True, title='S/N map', **kwargs)

    if verbose:
        print("S/N map created using {} processes".format(nproc))
        timing(start_time)
    return snrmap_array


def _snr_approx(array, source_xy, fwhm, centery, centerx):
    """
    array - frame convolved with top hat kernel
    """
    sourcex, sourcey = source_xy
    rad = dist(centery, centerx, sourcey, sourcex)
    ind_aper = disk((sourcey, sourcex), fwhm/2.)
    # noise : STDDEV in convolved array of 1px wide annulus (while
    # masking the flux aperture) * correction of # of resolution elements
    ind_ann = circle_perimeter(int(centery), int(centerx), int(rad))
    array2 = array.copy()
    array2[ind_aper] = mad(array[ind_ann])  # mask
    n2 = (2 * np.pi * rad) / fwhm - 1
    noise = array2[ind_ann].std(ddof=1) * np.sqrt(1+(1/n2))
    # signal : central px minus the mean of the pxs (masked) in 1px annulus
    signal = array[sourcey, sourcex] - array2[ind_ann].mean()
    snr_value = signal / noise
    return sourcey, sourcex, snr_value


def snr(array, source_xy, fwhm, full_output=False, array2=None, use2alone=False,
        exclude_negative_lobes=False, plot=False, verbose=False):
    """
    Calculate the S/N (signal to noise ratio) of a test resolution element
    in a residual frame (e.g. post-processed with LOCI, PCA, etc). Implements
    the approach described in Mawet et al. 2014 on small sample statistics,
    where a student t-test (eq. 9) can be used to determine S/N (and contrast)
    in high contrast imaging. 3 extra possibilities compared to Mawet et al. 
    2014 (https://ui.adsabs.harvard.edu/abs/2014ApJ...792...97M/abstract):

        * possibility to provide a second array (e.g. obtained with opposite
        derotation angles) to have more apertures for noise estimation
        * possibility to exclude negative ADI lobes directly adjacent to the
        tested xy location, to not bias the noise estimate
        * possibility to use only the second array for the noise estimation
        (useful for images containing a lot of disk/extended signals).
        
    *** DISCLAIMER ***
    Signal-to-noise ratio is not significance! For a conversion from snr to 
    n-sigma (i.e. the equivalent confidence level of a Gaussian n-sigma), use 
    the significance() function.    
    
    Parameters
    ----------
    array : numpy ndarray, 2d
        Post-processed frame where we want to measure S/N.
    source_xy : tuple of floats
        X and Y coordinates of the planet or test speckle.
    fwhm : float
        Size in pixels of the FWHM.
    full_output : bool, optional
        If True returns back the S/N value, the y, x input coordinates, noise
        and flux.
    array2 : None or numpy ndarray, 2d, optional
        Additional image (e.g. processed image with negative derotation angles) 
        enabling to have more apertures for noise estimation at each radial
        separation. Should have the same dimensions as array.
    use2alone : bool, opt
        Whether to use array2 alone to estimate the noise (can be useful to 
        estimate the S/N of extended disk features)
    exclude_negative_lobes : bool, opt
        Whether to include the adjacent aperture lobes to the tested location 
        or not. Can be set to True if the image shows significant neg lobes.
    plot : bool, optional
        Plots the frame and the apertures considered for clarity. 
    verbose: bool, optional
        Chooses whether to print some output or not.

    Returns
    -------
    [if full_output=True:]
    sourcey : numpy ndarray
        [full_output=True] Input coordinates (``source_xy``) in Y.
    sourcex : numpy ndarray
        [full_output=True] Input coordinates (``source_xy``) in X.
    f_source : float
        [full_output=True] Flux in test elemnt.
    fluxes : numpy ndarray
        [full_output=True] Background apertures fluxes.
    [always:]    
    snr_vale : float
        Value of the S/N for the given test resolution element.
    """
    check_array(array, dim=2, msg='array')
    if not isinstance(source_xy, tuple):
        raise TypeError("`source_xy` must be a tuple of floats")
    if array2 is not None:
        if not array2.shape == array.shape:
            raise TypeError('`array2` has not the same shape as input array')

    sourcex, sourcey = source_xy
    centery, centerx = frame_center(array)
    sep = dist(centery, centerx, float(sourcey), float(sourcex))

    if not sep > (fwhm/2)+1:
        raise RuntimeError('`source_xy` is too close to the frame center')

    sens = 'clock'  # counterclock

    angle = np.arcsin(fwhm/2./sep)*2
    number_apertures = int(np.floor(2*np.pi/angle))
    yy = np.zeros((number_apertures))
    xx = np.zeros((number_apertures))
    cosangle = np.cos(angle)
    sinangle = np.sin(angle)
    xx[0] = sourcex - centerx
    yy[0] = sourcey - centery
    for i in range(number_apertures-1):
        if sens == 'clock':
            xx[i+1] = cosangle*xx[i] + sinangle*yy[i] 
            yy[i+1] = cosangle*yy[i] - sinangle*xx[i] 
        elif sens == 'counterclock':
            xx[i+1] = cosangle*xx[i] - sinangle*yy[i] 
            yy[i+1] = cosangle*yy[i] + sinangle*xx[i]

    xx += centerx
    yy += centery
    rad = fwhm/2.
    if exclude_negative_lobes:
        xx = np.concatenate(([xx[0]], xx[2:-1]))
        yy = np.concatenate(([yy[0]], yy[2:-1]))

    apertures = photutils.CircularAperture((xx, yy), r=rad)  # Coordinates (X,Y)
    fluxes = photutils.aperture_photometry(array, apertures, method='exact')
    fluxes = np.array(fluxes['aperture_sum'])

    if array2 is not None:
        fluxes2 = photutils.aperture_photometry(array2, apertures,
                                                method='exact')
        fluxes2 = np.array(fluxes2['aperture_sum'])
        if use2alone:
            fluxes = np.concatenate(([fluxes[0]], fluxes2[:]))
        else:
            fluxes = np.concatenate((fluxes, fluxes2))

    f_source = fluxes[0].copy()
    fluxes = fluxes[1:]
    n2 = fluxes.shape[0]
    backgr_apertures_std = fluxes.std(ddof=1)
    snr_vale = (f_source - fluxes.mean())/(backgr_apertures_std *
                                           np.sqrt(1+(1/n2)))

    if verbose:
        msg1 = 'S/N for the given pixel = {:.3f}'
        msg2 = 'Integrated flux in FWHM test aperture = {:.3f}'
        msg3 = 'Mean of background apertures integrated fluxes = {:.3f}'
        msg4 = 'Std-dev of background apertures integrated fluxes = {:.3f}'
        print(msg1.format(snr_vale))
        print(msg2.format(f_source))
        print(msg3.format(fluxes.mean()))
        print(msg4.format(backgr_apertures_std))

    if plot:
        _, ax = plt.subplots(figsize=(6, 6))
        ax.imshow(array, origin='lower', interpolation='nearest', alpha=0.5, 
                  cmap='gray')
        for i in range(xx.shape[0]):
            # Circle takes coordinates as (X,Y)
            aper = plt.Circle((xx[i], yy[i]), radius=fwhm/2., color='r',
                              fill=False, alpha=0.8)
            ax.add_patch(aper)
            cent = plt.Circle((xx[i], yy[i]), radius=0.8, color='r', fill=True,
                              alpha=0.5)
            ax.add_patch(cent)
            aper_source = plt.Circle((sourcex, sourcey), radius=0.7,
                                     color='b', fill=True, alpha=0.5)
            ax.add_patch(aper_source)
        ax.grid(False)
        plt.show()

    if full_output:
        return sourcey, sourcex, f_source, fluxes, snr_vale
    else:
        return snr_vale



def significance(snr, rad, fwhm, student_to_gauss=True):
    """ Converts a S/N ratio (measured as in Mawet et al. 2014) into the 
    equivalent gaussian significance, i.e. the n-sigma with the same confidence 
    level as the S/N at the given separation.
     
     
    Parameters
    ----------
    snr : float or numpy array
        SNR value(s)
    rad : float or numpy array
        Radial separation(s) from the star in pixels. If an array, it should be
        the same shape as snr and provide the radial separation corresponding
        to each snr measurement.
    fwhm : float
        Full Width Half Maximum of the PSF.
    student_to_gauss : bool, optional
        Whether the conversion is from Student SNR to Gaussian significance. If 
        False, will assume the opposite: Gaussian significance to Student SNR.
    
    Returns
    -------
    sigma : float
        Gaussian significance in terms of n-sigma
    
    """
    
    if student_to_gauss:
        sigma = norm.ppf(t.cdf(snr,(rad/fwhm)*2*np.pi-2))
        if t.cdf(snr,(rad/fwhm)*2*np.pi-2)==1.0:
            print("Warning high S/N! cdf>0.9999999999999999 is rounded to 1")
            print(r"Returning 8.2 instead of inf. Quote detection>8.2 $\sigma$")
            return 8.2
    else:
        sigma = t.ppf(norm.cdf(snr), (rad/fwhm)*2*np.pi-2)
        
    return sigma      
            

def frame_report(array, fwhm, source_xy=None, verbose=True):
    """ Gets information from given frame: Integrated flux in aperture, S/N of
    central pixel (either ``source_xy`` or max value), mean S/N in aperture.

    Parameters
    ----------
    array : numpy ndarray
        2d array or input frame.
    fwhm : float
        Size of the FWHM in pixels.
    source_xy : tuple of floats or list (of tuples of floats)
        X and Y coordinates of the center(s) of the source(s).
    verbose : bool, optional
        If True prints to stdout the frame info.

    Returns
    -------
    source_xy : tuple of floats or list (of tuples of floats)
        X and Y coordinates of the center(s) of the source(s).
    obj_flux : list of floats
        Integrated flux in aperture.
    snr_centpx : list of floats
        S/N of the ``source_xy`` pixels.
    meansnr_pixels : list of floats
        Mean S/N of pixels in 1xFWHM apertures centered on ``source_xy``.
    """
    if array.ndim != 2:
        raise TypeError('Array is not 2d.')

    obj_flux = []
    meansnr_pixels = []
    snr_centpx = []

    if source_xy is not None:
        if isinstance(source_xy, (list, tuple)):
            if not isinstance(source_xy[0], tuple):
                source_xy = [source_xy]
        else:
            raise TypeError("`source_xy` must be a tuple of floats or tuple "
                            "of tuples")

        for xy in source_xy:
            x, y = xy
            if verbose:
                print(sep)
                print('Coords of chosen px (X,Y) = {:.1f}, {:.1f}'.format(x, y))

            # we get integrated flux on aperture with diameter=1FWHM
            aper = photutils.CircularAperture((x, y), r=fwhm / 2.)
            obj_flux_i = photutils.aperture_photometry(array, aper,
                                                       method='exact')
            obj_flux_i = obj_flux_i['aperture_sum'][0]

            # we get the mean and stddev of SNRs on aperture
            yy, xx = disk((y, x), fwhm / 2)
            snr_pixels_i = [snr(array, (x_, y_), fwhm, plot=False,
                                verbose=False) for y_, x_ in zip(yy, xx)]
            meansnr_i = np.mean(snr_pixels_i)
            stdsnr_i = np.std(snr_pixels_i,ddof=1)
            pxsnr_i = snr(array, (x, y), fwhm, plot=False, verbose=False)

            obj_flux.append(obj_flux_i)
            meansnr_pixels.append(meansnr_i)
            snr_centpx.append(pxsnr_i)

            if verbose:
                msg0 = 'Flux in a centered 1xFWHM circular aperture = {:.3f}'
                print(msg0.format(obj_flux_i))
                print('Central pixel S/N = {:.3f}'.format(pxsnr_i))
                print(sep)
                print('Inside a centered 1xFWHM circular aperture:')
                msg1 = 'Mean S/N (shifting the aperture center) = {:.3f}'
                print(msg1.format(meansnr_i))
                msg2 = 'Max S/N (shifting the aperture center) = {:.3f}'
                print(msg2.format(np.max(snr_pixels_i)))
                msg3 = 'stddev S/N (shifting the aperture center) = {:.3f}'
                print(msg3.format(stdsnr_i))
                print('')

    else:
        y, x = np.where(array == array.max())
        y, x = y[0], x[0]  # assuming there is only one max, taking 1st if clump
        source_xy = (x, y)
        if verbose:
            print(sep)
            print('Coords of Max px (X,Y) = {:.1f}, {:.1f}'.format(x, y))

        # we get integrated flux on aperture with diameter=1FWHM
        aper = photutils.CircularAperture((x, y), r=fwhm / 2.)
        obj_flux_i = photutils.aperture_photometry(array, aper, method='exact')
        obj_flux_i = obj_flux_i['aperture_sum'][0]

        # we get the mean and stddev of SNRs on aperture
        yy, xx = disk((y, x), fwhm / 2.)
        snr_pixels_i = [snr(array, (x_, y_), fwhm, plot=False,
                            verbose=False) for y_, x_ in zip(yy, xx)]
        meansnr_pixels = np.mean(snr_pixels_i)
        stdsnr_i = np.std(snr_pixels_i,ddof=1)
        pxsnr_i = snr(array, (x, y), fwhm, plot=False, verbose=False)

        obj_flux.append(obj_flux_i)
        snr_centpx.append(pxsnr_i)

        if verbose:
            msg0 = 'Flux in a centered 1xFWHM circular aperture = {:.3f}'
            print(msg0.format(obj_flux_i))
            print('Central pixel S/N = {:.3f}'.format(pxsnr_i))
            print(sep)
            print('Inside a centered 1xFWHM circular aperture:')
            msg1 = 'Mean S/N (shifting the aperture center) = {:.3f}'
            print(msg1.format(meansnr_pixels))
            msg2 = 'Max S/N (shifting the aperture center) = {:.3f}'
            print(msg2.format(np.max(snr_pixels_i)))
            msg3 = 'stddev S/N (shifting the aperture center) = {:.3f}'
            print(msg3.format(stdsnr_i))
            print(sep)

    return source_xy, obj_flux, snr_centpx, meansnr_pixels



#! /usr/bin/env python

"""
Implementation of the STIM map from [PAI18]
.. [PAI18]
   | Paret et al, 1018
   | **STIM map: detection map for exoplanets imaging beyond asymptotic Gaussian residual speckle noise**
   | *submitted in Monthly Notices of the Royal Astronomical Society*
"""
__author__ = 'Benoit Pairet'
__all__ = ['compute_stim_map',
           'compute_inverse_stim_map']

import numpy as np
from ..preproc import cube_derotate
from ..var import get_circle


def compute_stim_map(cube_der):
    """
    Computes the STIM detection map.

    Parameters
    ----------
    cube_der : 3d numpy ndarray
        Input de-rotated cube, e.g. output 'residuals_cube_' from
        ``vip_hci.pca.pca``.

    Returns
    -------
    detection_map : 2d ndarray
        STIM detection map.
    """
    t, n, _ = cube_der.shape
    mu = np.mean(cube_der, axis=0)
    sigma = np.sqrt(np.var(cube_der, axis=0))
    detection_map = np.divide(mu, sigma, out=np.zeros_like(mu),
                              where=sigma != 0)
    return get_circle(detection_map, int(np.round(n/2.)))


def compute_inverse_stim_map(cube, angle_list):
    """
    Computes the STIM detection map.

    Parameters
    ----------
    cube : 3d numpy ndarray
        Non de-rotated residuals from reduction algorithm, eg. output residuals
        from ``vip_hci.pca.pca``.
    angle_list : numpy ndarray, 1d
        Corresponding parallactic angle for each frame.    

    Returns
    -------
    inverse_stim_map : 2d ndarray
        Inverse STIM detection map.
    """
    t, n, _ = cube.shape
    cube_inv_der = cube_derotate(cube, -angle_list)
    inverse_stim_map = compute_stim_map(cube_inv_der)
    return inverse_stim_map


#! /usr/bin/env python

"""
Function defining the goodness of fit.
"""

__author__ = 'Valentin Christiaens'
__all__ = ['goodness_of_fit',
           'gof_scal']

import numpy as np
from matplotlib import pyplot as plt
from ..conf import vip_figsize
from .model_resampling import resample_model
from .utils_spec import extinction

def goodness_of_fit(lbda_obs, spec_obs, err_obs, lbda_mod, spec_mod, 
                    dlbda_obs=None, instru_corr=None, instru_fwhm=None, 
                    instru_idx=None, filter_reader=None, plot=False, 
                    outfile=None):
    """ Function to estimate the goodness of fit indicator defined as 
    in Olofsson et al. 2016 (Eq. 8). In addition, if a spectral 
    correlation matrix is provided, it is used to take into account the 
    correlated noise between spectral channels (see Greco & Brandt 2016).
    The goodness of fit indicator is identical to a chi square when all
    points are obtained from the same instrument (no additional weighting).

    Parameters
    ----------
    lbda_obs : numpy 1d ndarray or list
        Wavelength of observed spectrum. If several instruments, should be 
        ordered per instrument, not necessarily as monotonically increasing 
        wavelength. Hereafter, n_ch = len(lbda_obs).
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of lbda_obs.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. If 2d array, should be [2,n_ch]
        where the first (resp. second) column corresponds to lower (upper) 
        uncertainty, and n_ch is the length of lbda_obs and spec_obs.
    lbda_mod : numpy 1d ndarray or list
        Wavelength of tested model. Should have a wider wavelength extent than 
        the observed spectrum.
    spec_mod : numpy 1d ndarray
        Model spectrum. It does not require the same wavelength sampling as the
        observed spectrum. If higher spectral resolution, it will be convolved
        with the instrumental spectral psf (if instru_fwhm is provided) and 
        then binned to the same sampling. If lower spectral resolution, a 
        linear interpolation is performed to infer the value at the observed 
        spectrum wavelength sampling.
    dlbda_obs: numpy 1d ndarray or list, optional
        Spectral channel width for the observed spectrum. It should be provided 
        IF one wants to weigh each point based on the spectral 
        resolution of the respective instruments (as in Olofsson et al. 2016).
    instru_corr : numpy 2d ndarray or list, optional
        Spectral correlation throughout post-processed images in which the 
        spectrum is measured. It is specific to the combination of instrument, 
        algorithm and radial separation of the companion from the central star.
        Can be computed using distances.spectral_correlation(). In case of
        a spectrum obtained with different instruments, build it with
        distances.combine_corrs(). If not provided, it will consider the 
        uncertainties in each spectral channels are independent. See Greco & 
        Brandt (2017) for details.
    instru_fwhm : float or list, optional
        The instrumental spectral fwhm provided in nm. This is used to convolve
        the model spectrum. If several instruments are used, provide a list of 
        instru_fwhm values, one for each instrument whose spectral resolution
        is coarser than the model - including broad band
        filter FWHM if relevant.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to n_instru. Zero for points 
        that don't correspond to any instru_fwhm provided above, and i in 
        [1,n_instru] for points associated to instru_fwhm[i-1]. This parameter 
        must be provided if the spectrum consists of points obtained with 
        different instruments.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_fwhm, the default format assumed for the files:
        - first row containing header
        - starting from 2nd row: 1st column: WL in mu, 2nd column: transmission
        Note: files should all have the same format and wavelength units.
    plot : bool, optional
        Whether to plot the 
    outfile : string, optional
        Path+filename for the plot to be saved if provided.
        
    Returns
    -------
    chi_sq : float
        Goodness of fit indicator.

    """

    lbda_obs = np.array(lbda_obs)
    spec_obs = np.array(spec_obs)
    err_obs = np.array(err_obs)
    lbda_mod = np.array(lbda_mod)
    spec_mod = np.array(spec_mod)
    
    if lbda_obs.ndim != 1 or spec_obs.ndim != 1:
        raise TypeError('Observed lbda and spec must be lists or 1d arrays')
    if lbda_obs.shape[0] != spec_obs.shape[0]:
        raise TypeError('Obs lbda and spec need same shape')        
    if lbda_obs.shape[0] != err_obs.shape[-1]:
        raise TypeError('Obs lbda and err need same shape')
    if lbda_mod.ndim != 1 or spec_mod.ndim != 1:
        raise TypeError('The input model lbda/spec must be lists or 1d arrays')
    else:
        if lbda_mod.shape != spec_mod.shape:
            raise TypeError('The input model lbda and spec need same shape')
    if dlbda_obs is not None:
        if isinstance(dlbda_obs, list):
            dlbda_obs = np.array(dlbda_obs)
        if lbda_obs.shape != dlbda_obs.shape:
            raise TypeError('The input lbda_obs and dlbda_obs need same shape')

    n_ch = lbda_obs.shape[0]
             
    # interpolate OR convolve+bin model spectrum if not same sampling
    if not np.allclose(lbda_obs, lbda_mod):
        spec_mod_fin = resample_model(lbda_obs, lbda_mod, spec_mod, dlbda_obs, 
                                      instru_fwhm, instru_idx, filter_reader)    
    else:
        spec_mod_fin = spec_mod
        
    # compute covariance matrix
    if err_obs.ndim == 2:
        err_obs = (np.absolute(err_obs[0])+np.absolute(err_obs[1]))/2.
    if instru_corr is None:
        instru_corr = np.diag(np.ones(n_ch))
    cov = np.ones_like(instru_corr)
    for ii in range(n_ch):
        for jj in range(n_ch):
            cov[ii,jj] = instru_corr[ii,jj]*err_obs[ii]*err_obs[jj] 

    delta = spec_obs-spec_mod_fin
    wi = np.ones_like(dlbda_obs)
    if dlbda_obs is not None:
        if np.sum(np.power((dlbda_obs[1:]/lbda_obs[1:])-(dlbda_obs[:-1]/lbda_obs[:-1]),2))!=0:
            # normalize weights for their sum to be equal to the number of points
            wi = np.sqrt(((dlbda_obs/lbda_obs)/np.sum(dlbda_obs/lbda_obs))*dlbda_obs.shape[0])    
   
    chi_sq = np.linalg.multi_dot((wi*delta,np.linalg.inv(cov),wi*delta))
                
    if plot:
        _, ax = plt.subplots(figsize=vip_figsize)

        ax.plot(lbda_obs, lbda_obs*spec_obs, 'o', alpha=0.6, color='#1f77b4',
                label="Measured spectrum")
        ax.plot(lbda_obs, lbda_mod*spec_mod, '-', alpha=0.4, color='#1f77b4',
                label="Model")
        plt.xlabel('Wavelength')
        plt.ylabel(r'Flux density ($\lambda F_{\lambda}$)')

        plt.xlim(xmin=0.9*lbda_obs[0], xmax=1.1*lbda_obs[-1])
        plt.minorticks_on()
        plt.legend(fancybox=True, framealpha=0.5, fontsize=12, loc='best')
        plt.grid(which='major', alpha=0.2)
        if outfile:
            plt.savefig(outfile, bbox_inches='tight')
        plt.show()

    return chi_sq


def gof_scal(params, lbda_obs, spec_obs, err_obs, lbda_mod, spec_mod, dlbda_obs, 
             instru_corr, instru_fwhm, instru_idx, filter_reader, ext_range):
    """ Wrapper for the goodness of fit routine to search for best template 
    library fitting spectrum. The only difference is the "params" argument, 
    
    Parameters
    ----------
    params: tuple
        Tuple of 1 or 2 elements: scaling factor and (optionally) differential
        optical extinction (i.e. Delta A_V can be negative if template spectra
        are not dereddened).
    lbda_obs : numpy 1d ndarray or list
        Wavelength of observed spectrum. If several instruments, should be 
        ordered per instrument, not necessarily as monotonically increasing 
        wavelength. Hereafter, n_ch = len(lbda_obs).
    spec_obs : numpy 1d ndarray or list
        Observed spectrum for each value of lbda_obs.
    err_obs : numpy 1d/2d ndarray or list
        Uncertainties on the observed spectrum. If 2d array, should be [2,n_ch]
        where the first (resp. second) column corresponds to lower (upper) 
        uncertainty, and n_ch is the length of lbda_obs and spec_obs.
    lbda_mod : numpy 1d ndarray or list
        Wavelength of tested model. Should have a wider wavelength extent than 
        the observed spectrum.
    spec_mod : numpy 1d ndarray
        Model spectrum. It does not require the same wavelength sampling as the
        observed spectrum. If higher spectral resolution, it will be convolved
        with the instrumental spectral psf (if instru_fwhm is provided) and 
        then binned to the same sampling. If lower spectral resolution, a 
        linear interpolation is performed to infer the value at the observed 
        spectrum wavelength sampling.
    dlbda_obs: numpy 1d ndarray or list, optional
        Spectral channel width for the observed spectrum. It should be provided 
        IF one wants to weigh each point based on the spectral 
        resolution of the respective instruments (as in Olofsson et al. 2016).
    instru_corr : numpy 2d ndarray or list, optional
        Spectral correlation throughout post-processed images in which the 
        spectrum is measured. It is specific to the combination of instrument, 
        algorithm and radial separation of the companion from the central star.
        Can be computed using distances.spectral_correlation(). In case of
        a spectrum obtained with different instruments, build it with
        distances.combine_corrs(). If not provided, it will consider the 
        uncertainties in each spectral channels are independent. See Greco & 
        Brandt (2017) for details.
    instru_fwhm : float or list, optional
        The instrumental spectral fwhm provided in nm. This is used to convolve
        the model spectrum. If several instruments are used, provide a list of 
        instru_fwhm values, one for each instrument whose spectral resolution
        is coarser than the model - including broad band
        filter FWHM if relevant.
    instru_idx: numpy 1d array, optional
        1d array containing an index representing each instrument used 
        to obtain the spectrum, label them from 0 to n_instru. Zero for points 
        that don't correspond to any instru_fwhm provided above, and i in 
        [1,n_instru] for points associated to instru_fwhm[i-1]. This parameter 
        must be provided if the spectrum consists of points obtained with 
        different instruments.
    filter_reader: python routine, optional
        External routine that reads a filter file and returns a 2D numpy array, 
        where the first column corresponds to wavelengths, and the second 
        contains transmission values. Important: if not provided, but strings 
        are detected in instru_fwhm, the default format assumed for the files:
        - first row containing header
        - starting from 2nd row: 1st column: WL in mu, 2nd column: transmission
        Note: files should all have the same format and wavelength units.
    ext_range: tuple or None, opt
        If None: differential extinction is not to be considered as a free 
        parameter. Elif a tuple of 3 floats is provided, differential extinction 
        will be considered, with the floats as lower limit, upper limit and step 
        of the grid search.
        Note: if simplex search, the range is still used to set a chi of 
        np.inf outside of the range.
        
    Returns
    -------
    chi_sq : float
        Goodness of fit indicator.
    """
    tmp_spec = spec_mod*params[0]
    
    if len(params) == 2:
        if ext_range is None:
            raise TypeError("ext_range should be a tuple of 3 elements")
        else:
            if params[1]<ext_range[0] or params[1]>ext_range[1]:
                return np.inf
            AV=params[1]
            thr = ext_range[-1]
            if abs(AV) < thr:
                AV = 0
            Albdas = extinction(lbda_obs,abs(AV))
            extinc_fac = np.power(10.,-Albdas/2.5)
            if AV>0:
                tmp_spec *= extinc_fac
            elif AV<0:
                tmp_spec /= extinc_fac
    elif len(params) > 2:
        raise TypeError("params tuple should have length 1 or 2")
    
    
    return goodness_of_fit(lbda_obs, spec_obs, err_obs, lbda_obs, tmp_spec, 
                           dlbda_obs=dlbda_obs, instru_corr=instru_corr, 
                           instru_fwhm=instru_fwhm, instru_idx=instru_idx, 
                           filter_reader=filter_reader)


#! /usr/bin/env python
"""
Scattered_light_disk class definition
"""

__author__ = 'Julien Milli'
__all__ = []

import numpy as np
from ..var import frame_center
from .dust_distribution import Dust_distribution
from .phase_function import Phase_function


class ScatteredLightDisk(object):
    """
    Class used to generate a synthetic disc, inspired from a light version of 
    the GRATER tool (GRenoble RAdiative TransfER, Augereau et al. 1999) 
    written originally in IDL and converted to Python by J. Milli.
    """
    
    def __init__(self, nx=200, ny=200, distance=50., itilt=60., omega=0.,
                 pxInArcsec=0.01225, pa=0., flux_max=None,
                 density_dico={'name': '2PowerLaws', 'ain':5, 'aout':-5,
                               'a':40, 'e':0, 'ksi0':1., 'gamma':2., 'beta':1.,\
                                'dens_at_r0':1.},
                 spf_dico={'name':'HG', 'g':0., 'polar':False}, xdo=0., ydo=0.):
        """
        Constructor of the Scattered_light_disk object, taking in input the 
        geometric parameters of the disk, the radial density distribution 
        and the scattering phase function.
        So far, only one radial distribution is implemented: a smoothed
        2-power-law distribution, but more complex radial profiles can be implemented
        on demand.
        The star is assumed to be centered at the frame center as defined in
        the vip_hci.var.frame_center function (geometric center of the image,
        e.g. either in the middle of the central pixel for odd-size images or 
        in between 4 pixel for even-size images).
 
        Parameters
        ----------
        nx : int
            number of pixels along the x axis of the image (default 200)
        ny : int
            number of pixels along the y axis of the image (default 200)
        distance : float
            distance to the star in pc (default 70.)
        itilt : float
            inclination wrt the line of sight in degrees (0 means pole-on,
            90 means edge-on, default 60 degrees)
        omega : float
            argument of the pericenter in degrees (0 by default) 
        pxInArcsec : float 
            pixel field of view in arcsec/px (default the SPHERE pixel 
            scale 0.01225 arcsec/px)
        pa : float
            position angle of the disc in degrees (default 0 degrees, e.g. North)
        flux_max : float
            the max flux of the disk in ADU. By default None, meaning that
            the disk flux is not normalized to any value.
        density_dico : dict
            Parameters describing the dust density distribution function 
            to be implemented. By default, it uses a two-power law dust
            distribution with a vertical gaussian distribution with 
            linear flaring. This dictionary should at least contain the key
            "name".
            For a to-power law distribution, you can set it with
            'name:'2PowerLaws' and with the following parameters:
                a : float 
                    reference radius in au (default 40)
                ksi0 : float
                    scale height in au at the reference radius (default 1 a.u.)
                gamma : float 
                    exponent (2=gaussian,1=exponential profile, default 2)
                beta : float 
                    flaring index (0=no flaring, 1=linear flaring, default 1)
                ain : float 
                    slope of the power-low distribution in the inner disk. It 
                    must be positive (default 5)
                aout : float 
                    slope of the power-low distribution in the outer disk. It 
                    must be negative (default -5)
                e : float 
                    eccentricity (default 0)
                amin: float 
                    minimim semi-major axis: the dust density is 0 below this 
                    value (default 0)
        spf_dico :  dictionnary
            Parameters describing the scattering phase function to be implemented.
            By default, an isotropic phase function is implemented. It should
            at least contain the key "name". 
        xdo : float 
            disk offset along the x-axis in the disk frame (=semi-major axis), 
            in a.u. (default 0)
        ydo : float 
            disk offset along the y-axis in the disk frame (=semi-minor axis), 
            in a.u. (default 0)           
        """
        self.nx = nx    # number of pixels along the x axis of the image
        self.ny = ny    # number of pixels along the y axis of the image
        self.distance = distance  # distance to the star in pc
        self.set_inclination(itilt)
        self.set_omega(omega)
        self.set_flux_max(flux_max)
        self.pxInArcsec = pxInArcsec  # pixel field of view in arcsec/px
        self.pxInAU = self.pxInArcsec*self.distance     # 1 pixel in AU
        # disk offset along the x-axis in the disk frame (semi-major axis), AU
        self.xdo = xdo
        # disk offset along the y-axis in the disk frame (semi-minor axis), AU
        self.ydo = ydo
        self.rmin = np.sqrt(self.xdo**2+self.ydo**2)+self.pxInAU 
        self.dust_density = Dust_distribution(density_dico)     
        # star center along the y- and x-axis, in pixels
        self.yc, self.xc = frame_center(np.ndarray((self.ny,self.nx)))
        self.x_vector = (np.arange(0, nx) - self.xc)*self.pxInAU  # x axis in au
        self.y_vector = (np.arange(0, ny) - self.yc)*self.pxInAU  # y axis in au
        self.x_map_0PA, self.y_map_0PA = np.meshgrid(self.x_vector,
                                                     self.y_vector)
        self.set_pa(pa)
        self.phase_function = Phase_function(spf_dico=spf_dico)
        self.scattered_light_map = np.ndarray((ny, nx))
        self.scattered_light_map.fill(0.)
        
    def set_inclination(self, itilt):
        """
        Sets the inclination of the disk.

        Parameters
        ----------
        itilt : float 
            inclination of the disk wrt the line of sight in degrees (0 means 
            pole-on, 90 means edge-on, default 60 degrees)      
        """
        self.itilt = float(itilt)  # inclination wrt the line of sight in deg
        self.cosi = np.cos(np.deg2rad(self.itilt))
        self.sini = np.sin(np.deg2rad(self.itilt))
        
    def set_pa(self, pa):
        """
        Sets the disk position angle

        Parameters
        ----------
        pa : float
            position angle in degrees
        """
        self.pa = pa    # position angle of the disc in degrees
        self.cospa = np.cos(np.deg2rad(self.pa))
        self.sinpa = np.sin(np.deg2rad(self.pa))
        # rotation to get the disk major axis properly oriented, x in AU
        self.y_map = (self.cospa*self.x_map_0PA + self.sinpa*self.y_map_0PA)
        # rotation to get the disk major axis properly oriented, y in AU
        self.x_map = (-self.sinpa*self.x_map_0PA + self.cospa*self.y_map_0PA)
        
    def set_omega(self, omega):
        """
        Sets the argument of pericenter 

        Parameters
        ----------
        omega : float
            angle in degrees
        """
        self.omega = float(omega)

    def set_flux_max(self, flux_max):
        """
        Sets the mas flux of the disk

        Parameters
        ----------
        flux_max : float
            the max flux of the disk in ADU
        """
        self.flux_max = flux_max

    def set_density_distribution(self, density_dico):
        """
        Sets or updates the parameters of the density distribution    

        Parameters
        ----------
        density_dico : dict
            Parameters describing the dust density distribution function 
            to be implemented. By default, it uses a two-power law dust
            distribution with a vertical gaussian distribution with 
            linear flaring. This dictionary should at least contain the key
            "name". For a to-power law distribution, you can set it with
            name:'2PowerLaws' and with the following parameters:
            a : float 
                reference radius in au (default 60)
            ksi0 : float
                scale height in au at the reference radius (default 1 a.u.)
            gamma : float 
                exponent (2=gaussian,1=exponential profile, default 2)
            beta : float 
                flaring index (0=no flaring, 1=linear flaring, default 1)
            ain : float 
                slope of the power-low distribution in the inner disk. It 
                must be positive (default 5)
            aout : float 
                slope of the power-low distribution in the outer disk. It 
                must be negative (default -5)
            e : float 
                eccentricity (default 0)
        """
        self.dust_density.set_density_distribution(density_dico)
    
    def set_phase_function(self, spf_dico):
        """
        Sets the phase function of the dust 

        Parameters
        ----------
        spf_dico :  dict
            Parameters describing the scattering phase function to be
            implemented. Three phase functions are implemented so far: single
            Heyney Greenstein, double Heyney Greenstein and custum phase
            functions through interpolation. Read the constructor of each of
            those classes to know which parameters must be set in the dictionary
            in each case.
        """
        self.phase_function = Phase_function(spf_dico=spf_dico)
    
    def print_info(self):
        """ 
        Prints the information of the disk and image parameters
        """
        print('-----------------------------------')
        print('Geometrical properties of the image')
        print('-----------------------------------')
        print('Image size: {0:d} px by {1:d} px'.format(self.nx,self.ny))
        msg1 = 'Pixel size: {0:.4f} arcsec/px or {1:.2f} au/px'
        print(msg1.format(self.pxInArcsec, self.pxInAU))
        msg2 = 'Distance of the star {0:.1f} pc'
        print(msg2.format(self.distance))
        msg3 = 'From {0:.1f} au to {1:.1f} au in X'
        print(msg3.format(self.x_vector[0],self.x_vector[self.nx-1]))
        msg4 = 'From {0:.1f} au to {1:.1f} au in Y'
        print(msg4.format(self.y_vector[0], self.y_vector[self.nx-1]))
        print('Position angle of the disc: {0:.2f} degrees'.format(self.pa))
        print('Inclination {0:.2f} degrees'.format(self.itilt))
        print('Argument of pericenter {0:.2f} degrees'.format(self.omega))
        if self.flux_max is not None:
            print('Maximum flux of the disk {0:.2f}'.format(self.flux_max))
        self.dust_density.print_info()
        self.phase_function.print_info()
        
    def check_inclination(self):
        """
        Checks whether the inclination set is close to edge-on and risks to 
        induce artefacts from the limited numerical accuracy. In such a case
        the inclination is changed to be less edge-on.
        """
        if np.abs(np.mod(self.itilt,180)-90) < np.abs(np.mod(self.dust_density.dust_distribution_calc.itiltthreshold,180)-90):
            print('Warning the disk is too close to edge-on')
            msg = 'The inclination was changed from {0:.2f} to {1:.2f}'
            print(msg.format(self.itilt,
                             self.dust_density.dust_distribution_calc.itiltthreshold))
            self.set_inclination(self.dust_density.dust_distribution_calc.itiltthreshold)
    
    def compute_scattered_light(self, halfNbSlices=25):
        """ 
        Computes the scattered lignt image of the disk.

        Parameters
        ----------
        halfNbSlices : integer 
            half number of distances along the line of sight l
        """        
        self.check_inclination()
        # dist along the line of sight to reach the disk midplane (z_D=0), AU:
        lz0_map = self.y_map * np.tan(np.deg2rad(self.itilt))
        # dist to reach +zmax, AU:
        lzp_map = self.dust_density.dust_distribution_calc.zmax/self.cosi + \
                  lz0_map
        # dist to reach -zmax, AU:
        lzm_map = -self.dust_density.dust_distribution_calc.zmax/self.cosi + \
                  lz0_map
        dl_map = np.absolute(lzp_map-lzm_map)  # l range, in AU
        # squared maximum l value to reach the outer disk radius, in AU^2:
        lmax2 = self.dust_density.dust_distribution_calc.rmax**2 - (self.x_map**2+self.y_map**2)
        # squared minimum l value to reach the inner disk radius, in AU^2:
        lmin2 = (self.x_map**2+self.y_map**2)-self.rmin**2 
        validPixel_map = (lmax2 > 0.) * (lmin2 > 0.)        
        lwidth = 100.  # control the distribution of distances along l
        nbSlices = 2*halfNbSlices-1  # total number of distances 
                                     # along the line of sight
        tmp = (np.exp(np.arange(halfNbSlices)*np.log(lwidth+1.) /
                      (halfNbSlices-1.))-1.)/lwidth  # between 0 and 1
        ll = np.concatenate((-tmp[:0:-1], tmp))
        # 1d array pre-calculated values, AU
        ycs_vector = self.cosi*self.y_map[validPixel_map] 
        # 1d array pre-calculated values, AU
        zsn_vector = -self.sini*self.y_map[validPixel_map]  
        xd_vector = self.x_map[validPixel_map]  # x_disk, in AU
        limage = np.ndarray((nbSlices, self.ny, self.nx))
        limage.fill(0.)
        
        for il in range(nbSlices):
            # distance along the line of sight to reach the plane z
            l_vector =lz0_map[validPixel_map] + ll[il]*dl_map[validPixel_map] 
            # rotation about x axis 
            yd_vector = ycs_vector + self.sini * l_vector  # y_Disk in AU
            zd_vector = zsn_vector + self.cosi * l_vector  # z_Disk, in AU
            # Dist and polar angles in the frame centered on the star position:
            # squared distance to the star, in AU^2
            d2star_vector = xd_vector**2+yd_vector**2+zd_vector**2
            dstar_vector = np.sqrt(d2star_vector)  # distance to the star, in AU
            # midplane distance to the star (r coordinate), in AU
            rstar_vector = np.sqrt(xd_vector**2+yd_vector**2)
            thetastar_vector = np.arctan2(yd_vector, xd_vector)
            # Phase angles:
            cosphi_vector = (rstar_vector*self.sini*np.sin(thetastar_vector)+zd_vector*self.cosi)/dstar_vector  # in radians
            # Polar coordinates in the disk frame, and semi-major axis:
            # midplane distance to the disk center (r coordinate), in AU
            r_vector = np.sqrt((xd_vector-self.xdo)**2+(yd_vector-self.ydo)**2)
            # polar angle in radians between 0 and pi
            theta_vector = np.arctan2(yd_vector-self.ydo,xd_vector-self.xdo)
            costheta_vector = np.cos(theta_vector-np.deg2rad(self.omega))
            # Scattered light:
            # volume density
            rho_vector = self.dust_density.density_cylindrical(r_vector,
                                                               costheta_vector,
                                                               zd_vector)
            phase_function = self.phase_function.compute_phase_function_from_cosphi(cosphi_vector)
            image = np.ndarray((self.ny, self.nx))
            image.fill(0.)
            image[validPixel_map] = rho_vector*phase_function/d2star_vector
            limage[il,:,:] = image
        self.scattered_light_map.fill(0.)
        for il in range(1,nbSlices):
            self.scattered_light_map += (ll[il]-ll[il-1]) * (limage[il-1,:,:] +
                                                             limage[il,:,:])
        if self.flux_max is not None:
            self.scattered_light_map *= (self.flux_max/np.nanmax(self.scattered_light_map))
        return self.scattered_light_map


if __name__ == '__main__':
    """
    It is an example of use of the ScatteredLightDisk class. 
    """    

    # Example 1: creation of a disk at 20pc, with semi-major axis 20 a.u.
    #            inner and outer slopes 2 and -5, inclined by 70degrees wrt
    #            the line of sight and with a PA of 45degrees, an isotropic
    #            phase function
    Scattered_light_disk1 = ScatteredLightDisk(nx=201, ny=201, distance=20,
                                               itilt=70.,omega=0.,
                                               pxInArcsec=0.01225, pa=45.,
                                               density_dico={'name':'2PowerLaws',
                                                             'ain':5,'aout':-5,
                                                             'a':40,'e':0,
                                                             'ksi0':1.,
                                                             'gamma':2.,
                                                             'beta':1.},
                                               spf_dico={'name':'HG','g':0.,
                                                         'polar':False})
    disk1 = Scattered_light_disk1.compute_scattered_light()

    # If you prefer set the parameters in differen steps, you can also do that
    # with: (this is identical)
    Scattered_light_disk1 = ScatteredLightDisk(nx=201, ny=201, distance=20)
    Scattered_light_disk1.set_pa(45)
    Scattered_light_disk1.set_inclination(70)
    Scattered_light_disk1.set_density_distribution({'name':'2PowerLaws',
                                                    'ain':2,'aout':-5,'a':20,
                                                    'e':0,'ksi0':1.,
                                                    'gamma':2.,'beta':1.})
    Scattered_light_disk1.set_phase_function({'name':'HG','g':0.,'polar':False})
    disk1 = Scattered_light_disk1.compute_scattered_light()    

    # If you want to know what are the parameters you set for your disk:
    Scattered_light_disk1.print_info()

    # Example 2: Creation of a disk similar to example 1 but with a double 
    #           Heyney Greenstein phase function 

    Scattered_light_disk2 = ScatteredLightDisk(nx=201,ny=201,distance=20)
    Scattered_light_disk2.set_pa(45)
    Scattered_light_disk2.set_inclination(70)
    Scattered_light_disk2.set_density_distribution({'name':'2PowerLaws',
                                                    'ain':2,'aout':-5,'a':20,
                                                    'e':0,'ksi0':1.,
                                                    'gamma':2.,'beta':1.})
    Scattered_light_disk2.set_phase_function({'name':'DoubleHG',
                                              'g':[0.6,-0.6],'weight':0.7,
                                              'polar':False})
    Scattered_light_disk2.print_info()
    disk2 = Scattered_light_disk2.compute_scattered_light()    

    # Let's turn the polarisation on now:
    Scattered_light_disk2.set_phase_function({'name':'DoubleHG',
                                              'g':[0.6,-0.6],'weight':0.7,
                                              'polar':True})
    disk2_polar = Scattered_light_disk2.compute_scattered_light()    




"""
Subpackage ``stats`` contains functionalities such as:

- extracting statistics (mean, median, std dev, sum) in regions of a frame
  or cube,
- median absolute deviation,
- sigma filtering of pixels in frames,
- distance (correlation) between the frames in a cube,
- distance (correlation) between a cube and a reference frame.

"""
from .clip_sigma import *
from .distances import *
from .im_stats import *
from .utils_stats import *
